import os
import uuid

from django.conf import settings


def pytest_assertrepr_compare(config, op, left, right):
    """
    More details at: https://stackoverflow.com/a/50625086/3899136
    """
    if op in ("==", "!="):
        return ["{0} {1} {2}".format(left, op, right)]


def pytest_configure():
    settings.configure(
        INSTALLED_APPS=["django_stomp", "tests.support"],
        STOMP_SERVER_HOST=os.getenv("STOMP_SERVER_HOST"),
        STOMP_SERVER_PORT=os.getenv("STOMP_SERVER_PORT"),
        STOMP_SERVER_STANDBY_HOST=os.getenv("STOMP_SERVER_STANDBY_HOST"),
        STOMP_SERVER_STANDBY_PORT=os.getenv("STOMP_SERVER_STANDBY_PORT"),
        STOMP_SERVER_USER=os.getenv("STOMP_SERVER_USER"),
        STOMP_SERVER_PASSWORD=os.getenv("STOMP_SERVER_PASSWORD"),
        STOMP_USE_SSL=os.getenv("STOMP_USE_SSL"),
        STOMP_LISTENER_CLIENT_ID=os.getenv("STOMP_LISTENER_CLIENT_ID"),
        STOMP_CORRELATION_ID_REQUIRED=os.getenv("STOMP_CORRELATION_ID_REQUIRED"),
        STOMP_PROCESS_MSG_ON_BACKGROUND=os.getenv("STOMP_PROCESS_MSG_ON_BACKGROUND"),
        STOMP_OUTGOING_HEARTBEAT=os.getenv("STOMP_OUTGOING_HEARTBEAT"),
        STOMP_INCOMING_HEARTBEAT=os.getenv("STOMP_INCOMING_HEARTBEAT"),
        DATABASES={
            "default": {
                "ENGINE": os.getenv("DB_ENGINE", "django.db.backends.sqlite3"),
                "NAME": os.getenv("DB_DATABASE", f"test_db-{uuid.uuid4()}"),
                "USER": os.getenv("DB_USER"),
                "HOST": os.getenv("DB_HOST"),
                "PORT": os.getenv("DB_PORT"),
                "PASSWORD": os.getenv("DB_PASSWORD"),
                "TEST": {"NAME": os.getenv("DB_DATABASE", f"test_db-{uuid.uuid4()}")},
            }
        },
    )



import json
import logging
import urllib.parse
from time import sleep
from typing import Generator
from typing import Optional

import requests
from requests.adapters import HTTPAdapter

from tests.support.dtos import ConsumerStatus
from tests.support.dtos import CurrentDestinationStatus
from tests.support.dtos import MessageStatus

logger = logging.getLogger(__name__)

_queues_details_request_path = "/api/queues"
_specific_queue_details_request_path = _queues_details_request_path + "/%2F/{queue_name}"
_bindings_from_queue_request_path = _queues_details_request_path + "/%2F/{queue_name}/bindings"
_get_message_from_queue_request_path = _queues_details_request_path + "/%2F/{queue_name}/get"
_channels_details_request_path = "/api/channels"
_channel_details_from_channel_request_path = _channels_details_request_path + "/{channel_name}"
_overview_request_path = "/api/overview"


def current_queue_configuration(queue_name, host="localhost", port=15672) -> Optional[CurrentDestinationStatus]:
    result = _do_request(host, port, _specific_queue_details_request_path.format(queue_name=queue_name))

    logger.debug("RabbitMQ request result: %s", result)
    if result.get("error"):
        return None

    if result.get("message_stats"):
        message_stats = result["message_stats"]
        messages_dequeued = message_stats.get("deliver_get", 0)
        messages_enqueued = message_stats.get("publish")
    else:
        messages_dequeued = 0
        messages_enqueued = None

    number_of_pending_messages = result["messages"]
    number_of_consumers = result["consumers"]

    return CurrentDestinationStatus(
        number_of_pending_messages, number_of_consumers, messages_enqueued, messages_dequeued
    )


def current_topic_configuration(topic_name, host="localhost", port=15672) -> Optional[CurrentDestinationStatus]:
    queues = _do_request(host, port, _queues_details_request_path + "?name=&use_regex=false")
    for queue_details in queues:
        queue_name = queue_details["name"]
        bindings = _do_request(host, port, _bindings_from_queue_request_path.format(queue_name=queue_name))
        for binding in bindings:
            if binding["source"] == "amq.topic" and binding["routing_key"] == topic_name:
                message_stats = queue_details["message_stats"]
                number_of_pending_messages = queue_details["messages"]
                number_of_consumers = queue_details["consumers"]
                messages_enqueued = message_stats["publish"]
                messages_dequeued = message_stats["deliver_get"] if message_stats.get("deliver_get") else 0
                return CurrentDestinationStatus(
                    number_of_pending_messages, number_of_consumers, messages_enqueued, messages_dequeued
                )
    return None


def consumers_details(connection_id, host="localhost", port=15672) -> Generator[ConsumerStatus, None, None]:
    channels = _do_request(host, port, _channels_details_request_path)
    for channel in channels:
        channel_name = channel["connection_details"]["name"]
        channel_details = _do_request(
            host,
            port,
            _channel_details_from_channel_request_path.format(
                channel_name=urllib.parse.quote(f"{channel_name} ") + "(1)"
            ),
        )
        if channel_details.get("consumer_details"):
            for consumer in channel_details["consumer_details"]:
                if consumer["consumer_tag"] == f"T_{connection_id}":
                    yield ConsumerStatus(
                        address_to_destination_details=None,
                        destination_name=consumer["queue"]["name"],
                        session_id=None,
                        enqueues=None,
                        dequeues=None,
                        dispatched=None,
                        dispatched_queue=None,
                        prefetch=consumer["prefetch_count"],
                        max_pending=channel_details["messages_unacknowledged"],
                        exclusive=consumer["exclusive"],
                        retroactive=None,
                    )


def retrieve_message_published(destination_name, host="localhost", port=15672) -> MessageStatus:
    body = json.dumps(
        {
            "vhost": "/",
            "name": destination_name,
            "truncate": "50000",
            "ackmode": "ack_requeue_false",
            "encoding": "auto",
            "count": "1",
        }
    )
    message_details = _do_request(
        host, port, _get_message_from_queue_request_path.format(queue_name=destination_name), do_post=True, body=body
    )
    assert len(message_details) == 1
    properties = message_details[0]["properties"]

    details = json.loads(message_details[0]["payload"])
    persistent = None
    correlation_id = properties["correlation_id"]
    headers = properties.pop("headers")

    return MessageStatus(None, details, persistent, correlation_id, {**headers, **properties})


def get_broker_version(host="localhost", port=15672) -> str:
    broker_overview = _do_request(host, port, _overview_request_path)
    return broker_overview["rabbitmq_version"]


def _do_request(host, port, request_path, do_post=False, body=None):
    sleep(2)
    session = requests.Session()
    session.mount("http://", HTTPAdapter(max_retries=3))
    address, auth = f"http://{host}:{port}{request_path}", ("guest", "guest")
    with session:
        if not do_post:
            data = session.get(address, auth=auth)
        else:
            data = session.post(address, auth=auth, data=body)
    return data.json()



import json
from uuid import uuid4

from django_stomp import builder
from django_stomp.services.consumer import StompFrame
from django_stomp.services.consumer import build_listener
from tests.support.helpers import get_active_threads_name_with_prefix


def test_should_create_at_most_the_defined_number_of_workers(mocker):
    mocker.patch("django_stomp.services.consumer.STOMP_PROCESS_MSG_WORKERS", 3)

    listener = build_listener(f"some-destination-{uuid4()}", should_process_msg_on_background=True)

    fake_frame = StompFrame(cmd="FAKE_CMD", headers={"message-id": "123"}, body=json.dumps({"someKey": 1}))

    listener.on_message(fake_frame)
    listener.on_message(fake_frame)
    listener.on_message(fake_frame)
    listener.on_message(fake_frame)

    workers_threads = get_active_threads_name_with_prefix(listener._subscription_id)
    assert len(workers_threads) == 3  # 3 workers
    listener.shutdown_worker_pool()


def test_should_clean_up_worker_pool():
    listener = build_listener(f"some-destination-{uuid4()}", should_process_msg_on_background=True)

    fake_frame = StompFrame(cmd="FAKE_CMD", headers={"message-id": "123"}, body=json.dumps({"someKey": 1}))

    listener.on_message(fake_frame)

    workers_threads_before_pool_shutdown = get_active_threads_name_with_prefix(listener._subscription_id)
    assert len(workers_threads_before_pool_shutdown) == 1  # only one worker

    listener.shutdown_worker_pool()

    workers_threads_after_pool_shutdown = get_active_threads_name_with_prefix(listener._subscription_id)
    assert len(workers_threads_after_pool_shutdown) == 0  # no active worker thread


def test_should_still_process_message_if_worker_pool_was_explicitly_shutdown():
    listener = build_listener(f"some-destination-{uuid4()}", should_process_msg_on_background=True)

    fake_frame = StompFrame(cmd="FAKE_CMD", headers={"message-id": "123"}, body=json.dumps({"someKey": 1}))

    listener.on_message(fake_frame)

    workers_threads_before_pool_shutdown = get_active_threads_name_with_prefix(listener._subscription_id)
    assert len(workers_threads_before_pool_shutdown) == 1  # only one worker
    listener.shutdown_worker_pool()

    workers_threads_after_pool_shutdown = get_active_threads_name_with_prefix(listener._subscription_id)
    assert len(workers_threads_after_pool_shutdown) == 0  # no active worker thread

    listener.on_message(fake_frame)

    workers_threads_before_pool_shutdown = get_active_threads_name_with_prefix(listener._subscription_id)
    assert len(workers_threads_before_pool_shutdown) == 1  # only one worker

    listener.shutdown_worker_pool()


def test_should_have_only_one_django_stomp_listener_even_if_set_listener_is_called_multiple_times():
    # Arrange - build listener to some arbirtrary queue
    django_stomp_listener = builder.build_listener(f"some-destination-{uuid4()}")

    # Act - in sucession, multiples _set_listener call to trigger _connection.set_listener
    django_stomp_listener._set_listener()
    django_stomp_listener._set_listener()
    django_stomp_listener._set_listener()
    django_stomp_listener._set_listener()
    django_stomp_listener._set_listener()

    # Assert - it has two listeners (protocol-listener, stomp.py defined and the one defined by django-stomp)
    assert len(django_stomp_listener._connection.transport.listeners) == 2
    assert django_stomp_listener._connection.get_listener("protocol-listener") is not None

    django_stomp_listener_id = django_stomp_listener._listener_id
    assert django_stomp_listener._connection.get_listener(django_stomp_listener_id) is not None



from django.core.management import CommandError
from django.core.management.base import BaseCommand
from django_stomp.execution import send_message_from_one_destination_to_another


class Command(BaseCommand):
    help = "Move messages from one destination to another"

    def add_arguments(self, parser):
        parser.add_argument("source_destination", type=str, help="Source destination used to consume messages")
        parser.add_argument(
            "target_destination", type=str, help="The target which will receive messages from consumed destination"
        )
        parser.add_argument(
            "broker_host_for_source_destination",
            type=str,
            nargs="?",
            help="Custom broker host only to get messages from source_destination",
        )
        parser.add_argument(
            "broker_port_for_source_destination",
            type=int,
            nargs="?",
            help="Custom broker port only to get messages from source_destination",
        )

    def handle(self, *args, **options):
        self.stdout.write(f"Provided parameters: {options}")

        source_destination = options.get("source_destination")
        target_destination = options.get("target_destination")
        broker_host_for_source_destination = options.get("broker_host_for_source_destination")
        broker_port_for_source_destination = options.get("broker_port_for_source_destination")

        if broker_host_for_source_destination and not broker_port_for_source_destination:
            raise CommandError("You should provide broker_port_for_source_destination option as well")

        self.stdout.write("Calling internal service to send messages to target")

        send_message_from_one_destination_to_another(
            source_destination,
            target_destination,
            custom_stomp_server_host=broker_host_for_source_destination,
            custom_stomp_server_port=broker_port_for_source_destination,
        )



from uuid import uuid4

from django.db import models


class Simple(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4, editable=False)
    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)



from time import sleep
from typing import Generator

import requests

from parsel import Selector
from tests.support.dtos import SubscriberSetup


def offline_durable_subscribers(host) -> Generator[SubscriberSetup, None, None]:
    sleep(1)
    result = requests.get(f"http://{host}:8161/admin/subscribers.jsp", auth=("admin", "admin"))
    selector = Selector(text=str(result.content))

    all_offline_subscribers = (
        selector.xpath("//h2[contains(text(),'Offline Durable Topic Subscribers')]/following-sibling::table")[0]
        .css("table tbody tr")
        .getall()
    )

    for index, column_details in enumerate(all_offline_subscribers):
        column_details_as_selector = Selector(text=column_details)
        client_id_request_path = column_details_as_selector.css("td a::attr(href)").get()
        address_to_subscriber_details = f"http://{host}:8161/admin/{client_id_request_path}"
        subscriber_id = client_id_request_path.split("connectionID=")[1]
        destination = column_details_as_selector.css("td:nth-child(4) span::text").getall()[1]
        pending_queue_size = int(column_details_as_selector.css("td:nth-child(6)::text").get())
        dispatched_queue_size = int(column_details_as_selector.css("td:nth-child(7)::text").get())
        dispatched_counter = int(column_details_as_selector.css("td:nth-child(8)::text").get())
        enqueue_counter = int(column_details_as_selector.css("td:nth-child(9)::text").get())
        dequeue_counter = int(column_details_as_selector.css("td:nth-child(10)::text").get())
        yield SubscriberSetup(
            address_to_subscriber_details,
            subscriber_id,
            destination,
            pending_queue_size,
            dispatched_queue_size,
            dispatched_counter,
            enqueue_counter,
            dequeue_counter,
        )



"""
Module with testing callbacks used for the tests.
"""
import logging
import threading
from time import sleep

from django import db

from django_stomp.builder import build_publisher
from django_stomp.services.consumer import Payload
from tests.support.models import Simple

callback_move_and_ack_path = "tests.support.callbacks_for_tests.callback_move_and_ack"
callback_standard_path = "tests.support.callbacks_for_tests.callback_standard"
callback_with_exception_path = "tests.support.callbacks_for_tests.callback_with_exception"
callback_with_sleep_three_seconds_while_heartbeat_thread_is_alive_path = (
    "tests.support.callbacks_for_tests.callback_with_sleep_three_seconds_while_heartbeat_thread_is_alive"
)
callback_with_logging_path = "tests.support.callbacks_for_tests.callback_with_logging"
callback_with_sleep_three_seconds_path = "tests.support.callbacks_for_tests.callback_with_sleep_three_seconds"
callback_with_another_log_message_path = "tests.support.callbacks_for_tests.callback_with_another_log_message"
callback_with_nack_path = "tests.support.callbacks_for_tests.callback_with_nack"
callback_with_explicit_db_connection_path = "tests.support.callbacks_for_tests.callback_with_explicit_db_connection"


def callback_move_and_ack(payload: Payload, destination: str):
    """
    Callback for moving to another destination. Requires start_processing() to be called
    with param_to_callback.
    """
    publisher = build_publisher()
    publisher.send(payload.body, destination, attempt=1)
    payload.ack()


def callback_standard(payload: Payload):
    """
    Standard callback with simple ack.
    """
    # Should dequeue the message
    payload.ack()


def callback_with_nack(payload: Payload):
    payload.nack()


def callback_with_exception(payload: Payload):
    raise Exception("Lambe Sal")


def callback_with_sleep_three_seconds_while_heartbeat_thread_is_alive(payload: Payload):
    while True:
        sleep(3)
        heartbeat_threads = filter(lambda thread: "StompHeartbeatThread" in thread.name, threading.enumerate())
        if all(not thread.is_alive() for thread in heartbeat_threads):
            break


def callback_with_logging(payload: Payload):
    logger = logging.getLogger(__name__)
    logger.info("I'll process the message: %s!", payload.body)
    payload.ack()


def callback_with_another_log_message(payload: Payload):
    logger = logging.getLogger(__name__)
    logger.info("%s is the message that I'll process!", payload.body)
    payload.ack()


def callback_with_sleep_three_seconds(payload: Payload):
    sleep(3)
    payload.ack()
    logger = logging.getLogger(__name__)
    logger.info("%s sucessfully processed!", payload.body)


def callback_with_explicit_db_connection(payload: Payload):
    """
    This callback do a DB interaction and set a new attribute (`db`) in the thread that it's running
    so that any test can have a direct access to a `DatabaseWrapper`[1] and make assertions with it. This
    new attribute was needed due to the way Django works estabilishing a new DB connection for each
    thread [2].

    Note: The `db.connections` object is a handler (with a dict-like syntax), see [3] for more information
    on it!

    [1] https://github.com/django/django/blob/ca9872905559026af82000e46cde6f7dedc897b6/django/db/backends/base/base.py#L26
    [2] https://docs.djangoproject.com/en/3.2/ref/databases/#caveats
    [3] https://github.com/django/django/blob/ca9872905559026af82000e46cde6f7dedc897b6/django/db/utils.py#L134
    """
    Simple.objects.create()

    thread = threading.current_thread()
    setattr(thread, "db", db.connections["default"])

    payload.ack()



class CorrelationIdNotProvidedException(Exception):
    pass


class DjangoStompImproperlyConfigured(Exception):
    pass


class DjangoStompIncorrectUse(Exception):
    """
    Raised when Django stomp has been invoked in a wrong manner such as
    less arguments than it needs, etc.
    """

    pass



import uuid

import pytest
from django.core.management import call_command

from django_stomp.builder import build_publisher


@pytest.mark.skip("This is done manually for now")
def test_should_send_from_one_broker_to_another(settings):
    fake_source = "/queue/jafar-destination"
    fake_target = "/queue/another-jafar"

    def _to_publish_to_source_broker():
        publisher_name_to_move = f"iago-{uuid.uuid4()}"
        with build_publisher(publisher_name_to_move).auto_open_close_connection() as publisher:
            with publisher.do_inside_transaction():
                sample_body = {"igu": "brick-breaker", "is_enabled": True}
                headers = {"custom-1": 1234, "custom-2": "aladdin"}
                publisher.send(sample_body, fake_source, headers)

    def _to_move_the_message_and_evaluate_result():
        fake_custom_broker_host = "localhost"
        fake_custom_broker_port = "61614"

        call_command("move_messages", fake_source, fake_target, fake_custom_broker_host, fake_custom_broker_port)

    settings.STOMP_SERVER_PORT = 61614
    _to_publish_to_source_broker()
    settings.STOMP_SERVER_PORT = 61613
    _to_move_the_message_and_evaluate_result()



import json
import uuid
from datetime import datetime

import pytest
from django_stomp.builder import build_publisher
from django_stomp.services.consumer import Acknowledgements
from stomp.listener import TestListener

test_destination = "/queue/my-test-destination"


@pytest.fixture
def setup_publisher_and_test_listener():
    subscription_id = str(uuid.uuid4())
    test_listener = TestListener()
    publisher = build_publisher()
    publisher.connection.set_listener(str(uuid.uuid4()), test_listener)
    # Start is needed to make `connection.subscribe` work properly
    publisher.start()
    publisher.connection.subscribe(destination=test_destination, id=subscription_id, ack=Acknowledgements.AUTO.value)

    yield publisher, (test_listener, subscription_id)

    publisher.close()


def test_should_publish_message_on_destination(setup_publisher_and_test_listener):
    publisher, listener_configuration = setup_publisher_and_test_listener
    listener, subscription_id = listener_configuration

    some_correlation_id = uuid.uuid4()
    some_header = {"correlation-id": some_correlation_id}
    some_body = {"keyOne": 1, "keyTwo": 2}

    publisher.send(some_body, test_destination, headers=some_header, attempt=1)

    listener.wait_for_message()
    received_message = listener.get_latest_message()

    assert received_message is not None
    received_header = received_message[0]
    assert received_header["correlation-id"] == str(some_correlation_id)
    assert received_header["destination"] == test_destination
    assert received_header["subscription"] == subscription_id
    assert received_header["message-id"] is not None
    assert received_header["content-type"] == "application/json;charset=utf-8"
    assert received_header["persistent"] == "true"
    received_body = json.loads(received_message[1])
    assert received_body == some_body
    #### ActiveMQ has it by default
    # assert received_header["priority"] == "4"
    #### ActiveMQ has it by default
    # timestamp_value = int(received_header["timestamp"])
    # converted_timestamp = datetime.utcfromtimestamp(timestamp_value / 1000.0)
    # assert converted_timestamp.date() == datetime.today().date()



import re

import pytest

from django_stomp.helpers import retry


def test_should_log_retry_attempts_on_warning_level(caplog):
    def raise_exception():
        raise Exception("Frango depenado")

    with pytest.raises(Exception, match="Frango depenado"):
        retry(raise_exception, attempt=3)

    tenacity_logs_regex = re.compile(r"Finished call to .* after .*, this was the .* time calling it.")
    tenacity_logs = [record for record in caplog.records if tenacity_logs_regex.match(record.message)]
    assert len(tenacity_logs) == 3
    assert all([log.levelname == "WARNING" for log in tenacity_logs])



import pytest

from django_stomp.exceptions import DjangoStompImproperlyConfigured
from django_stomp.settings import eval_as_int_if_provided_value_is_not_none_otherwise_none
from django_stomp.settings import eval_settings_otherwise_raise_exception


def test_should_raise_improperly_configured_when_settings_is_not_correct_configured(mocker):
    mocked_settings = mocker.patch("django_stomp.settings.django_settings")
    mocked_settings.STOMP_PROCESS_MSG_WORKERS = "abc"
    expected_exception_message = "STOMP_PROCESS_MSG_WORKERS is not valid!"

    with pytest.raises(DjangoStompImproperlyConfigured, match=expected_exception_message):
        eval_settings_otherwise_raise_exception(
            "STOMP_PROCESS_MSG_WORKERS", eval_as_int_if_provided_value_is_not_none_otherwise_none
        )

    mocked_settings.STOMP_PROCESS_MSG_WORKERS = {}
    with pytest.raises(DjangoStompImproperlyConfigured, match=expected_exception_message):
        eval_settings_otherwise_raise_exception(
            "STOMP_PROCESS_MSG_WORKERS", eval_as_int_if_provided_value_is_not_none_otherwise_none
        )

    mocked_settings.STOMP_PROCESS_MSG_WORKERS = []
    with pytest.raises(DjangoStompImproperlyConfigured, match=expected_exception_message):
        eval_settings_otherwise_raise_exception(
            "STOMP_PROCESS_MSG_WORKERS", eval_as_int_if_provided_value_is_not_none_otherwise_none
        )


def test_should_evaluate_settings_when_it_is_configured_as_expected(mocker):
    mocked_settings = mocker.patch("django_stomp.settings.django_settings")
    mocked_settings.STOMP_PROCESS_MSG_WORKERS = None

    evaluated_settings = eval_settings_otherwise_raise_exception(
        "STOMP_PROCESS_MSG_WORKERS", eval_as_int_if_provided_value_is_not_none_otherwise_none
    )
    assert evaluated_settings is None

    mocked_settings.STOMP_PROCESS_MSG_WORKERS = 123
    evaluated_settings = eval_settings_otherwise_raise_exception(
        "STOMP_PROCESS_MSG_WORKERS", eval_as_int_if_provided_value_is_not_none_otherwise_none
    )
    assert evaluated_settings == 123

    mocked_settings.STOMP_PROCESS_MSG_WORKERS = "3"
    evaluated_settings = eval_settings_otherwise_raise_exception(
        "STOMP_PROCESS_MSG_WORKERS", eval_as_int_if_provided_value_is_not_none_otherwise_none
    )
    assert evaluated_settings == 3



import logging
import uuid
from time import sleep
from typing import Optional

from django import db
from django.conf import settings
from django.utils.module_loading import import_string
from request_id_django_log import local_threading

from django_stomp.builder import build_listener
from django_stomp.builder import build_publisher
from django_stomp.exceptions import CorrelationIdNotProvidedException
from django_stomp.helpers import create_dlq_destination_from_another_destination
from django_stomp.helpers import eval_str_as_boolean
from django_stomp.helpers import get_listener_client_id
from django_stomp.helpers import get_subscription_destination
from django_stomp.helpers import is_destination_from_virtual_topic
from django_stomp.helpers import remove_key_from_dict
from django_stomp.services.consumer import Listener
from django_stomp.services.consumer import Payload

logger = logging.getLogger("django_stomp")

wait_to_connect = int(getattr(settings, "STOMP_WAIT_TO_CONNECT", 10))
durable_topic_subscription = eval_str_as_boolean(getattr(settings, "STOMP_DURABLE_TOPIC_SUBSCRIPTION", False))
listener_client_id = getattr(settings, "STOMP_LISTENER_CLIENT_ID", None)
is_correlation_id_required = eval_str_as_boolean(getattr(settings, "STOMP_CORRELATION_ID_REQUIRED", True))
should_process_msg_on_background = eval_str_as_boolean(getattr(settings, "STOMP_PROCESS_MSG_ON_BACKGROUND", True))
publisher_name = "django-stomp-another-target"


def start_processing(
    destination_name: str,
    callback_str: str,
    is_testing=False,
    testing_disconnect=True,
    param_to_callback=None,
    return_listener=False,
    execute_workaround_to_deal_with_rabbit_mq=True,
    broker_host_to_consume_messages: Optional[str] = None,
    broker_port_to_consume_messages: Optional[int] = None,
) -> Optional[Listener]:
    callback_function = import_string(callback_str)

    if execute_workaround_to_deal_with_rabbit_mq:
        _create_dlq_queue(destination_name)
        if is_destination_from_virtual_topic(destination_name):
            routing_key = get_subscription_destination(destination_name)
            _create_queue(destination_name, routing_key=routing_key)
            logger.info("Created/Refreshed queue to consume from topic in case of RabbitMQ...")

    client_id = get_listener_client_id(durable_topic_subscription, listener_client_id)

    listener = build_listener(
        destination_name,
        durable_topic_subscription,
        client_id=client_id,
        should_process_msg_on_background=should_process_msg_on_background,
        custom_stomp_server_host=broker_host_to_consume_messages,
        custom_stomp_server_port=broker_port_to_consume_messages,
    )

    def main_logic() -> Optional[Listener]:
        try:
            logger.info("Starting listener...")

            def _callback(payload: Payload) -> None:
                try:
                    db.close_old_connections()
                    local_threading.request_id = _get_or_create_correlation_id(payload.headers)

                    if param_to_callback:
                        callback_function(payload, param_to_callback)
                    else:
                        callback_function(payload)
                except BaseException as e:
                    logger.exception(f"A exception of type {type(e)} was captured during callback logic")
                    logger.warning("Trying to do NACK explicitly sending the message to DLQ...")
                    if listener.is_open():
                        payload.nack()
                        logger.warning("Done!")
                    raise e
                finally:
                    local_threading.request_id = None
                    db.close_old_connections()

            listener.start(_callback, wait_forever=is_testing is False)

            if is_testing is True:
                return listener
        except BaseException as e:
            logger.exception(f"A exception of type {type(e)} was captured during listener logic")
        finally:
            if is_testing is False:
                logger.info(f"Trying to close listener...")
                if listener.is_open():
                    listener.close()
                logger.info(f"Waiting {wait_to_connect} seconds before trying to connect again...")
                sleep(wait_to_connect)

    if is_testing is False:
        while True:
            main_logic()
    else:
        max_tries = 3
        tries = 0
        testing_listener = None
        while True:
            if tries == 0:
                testing_listener = main_logic()
                if return_listener:
                    return testing_listener
                tries += 1
            elif tries >= max_tries:
                if testing_disconnect is True:
                    testing_listener.close()
                break
            else:
                sleep(0.2)
                tries += 1


def send_message_from_one_destination_to_another(
    source_destination: str,
    target_destination: str,
    is_testing: bool = False,
    testing_disconnect: bool = True,
    return_listener: bool = False,
    custom_stomp_server_host: Optional[str] = None,
    custom_stomp_server_port: Optional[int] = None,
) -> Listener:
    callback_function = "django_stomp.execution._callback_send_to_another_destination"

    return start_processing(
        source_destination,
        callback_function,
        is_testing=is_testing,
        testing_disconnect=testing_disconnect,
        param_to_callback=target_destination,
        return_listener=return_listener,
        execute_workaround_to_deal_with_rabbit_mq=False,
        broker_host_to_consume_messages=custom_stomp_server_host,
        broker_port_to_consume_messages=custom_stomp_server_port,
    )


def clean_messages_on_destination_by_acking(
    source_destination: str, is_testing: bool = False, testing_disconnect: bool = True, return_listener: bool = False
) -> Listener:
    """
    Cleans a queue by acking all messages on it (no queue purging or deleting).
    """
    ack_only_callback_path = "django_stomp.execution._callback_for_cleaning_queues"

    return start_processing(
        source_destination,
        ack_only_callback_path,
        is_testing=is_testing,
        testing_disconnect=testing_disconnect,
        return_listener=return_listener,
        execute_workaround_to_deal_with_rabbit_mq=False,
    )


def _callback_for_cleaning_queues(payload: Payload):
    """
    Callback that just acks all messages on a queue for cleaning it.
    """
    headers = payload.headers
    body = payload.body
    logger.info("Acking the following headers: %s, and body: %s", headers, body)

    payload.ack()
    logger.info("Message has been removed!")


def _callback_send_to_another_destination(payload: Payload, target_destination):
    logger.info(f"Message received!")

    headers = payload.headers
    body = payload.body

    logger.info(f"Configured headers: {headers}")
    logger.info(f"Configured body: {body}")
    logger.info("Sending to target destination...")
    publisher_name_to_move = f"{publisher_name}-{uuid.uuid4()}"

    with build_publisher(publisher_name_to_move).auto_open_close_connection() as publisher:
        with publisher.do_inside_transaction():
            # Remove the message-id header in the SEND frame since in RabbitMQ we cannot
            # set it: https://www.rabbitmq.com/stomp.html#pear.hpos
            remove_key_from_dict(headers, "message-id")
            publisher.send(body, target_destination, headers)
            payload.ack()

    logger.info("The messages has been moved!")


def _create_queue(queue_name: str, durable_topic_subscription: bool = False, routing_key: Optional[str] = None):
    client_id = get_listener_client_id(durable_topic_subscription, listener_client_id)
    listener = build_listener(queue_name, durable_topic_subscription, client_id=client_id, routing_key=routing_key)
    listener.start(lambda payload: None, wait_forever=False)
    listener.close()


def _create_dlq_queue(destination_name: str):
    dlq_destination_name = create_dlq_destination_from_another_destination(destination_name)
    _create_queue(dlq_destination_name)
    logger.info("Created/Refreshed DLQ in case of RabbitMQ...")


def _get_or_create_correlation_id(headers: dict) -> str:
    if "correlation-id" in headers:
        return headers["correlation-id"]

    if not is_correlation_id_required:
        correlation_id = uuid.uuid4()
        logger.info(f"New correlation-id was generated {correlation_id}")
        return str(correlation_id)

    raise CorrelationIdNotProvidedException(headers)



import json
import re
import threading
from time import sleep
from typing import Dict
from uuid import uuid4

from django.core.serializers.json import DjangoJSONEncoder

from django_stomp.builder import build_listener
from django_stomp.builder import build_publisher
from django_stomp.helpers import clean_dict_with_falsy_or_strange_values
from django_stomp.helpers import create_dlq_destination_from_another_destination
from django_stomp.helpers import retry
from tests.support import rabbitmq
from tests.support.activemq.queue_details import current_queue_configuration
from tests.support.dtos import CurrentDestinationStatus


def get_destination_metrics_from_broker(destination_name: str) -> CurrentDestinationStatus:
    """
    Retrieves metrics from the broker management console regardless of the broker engine
    ActiveMQ or RabbitMQ.
    """
    try:
        destination_metrics = current_queue_configuration(destination_name)
    except Exception:
        destination_metrics = rabbitmq.current_queue_configuration(destination_name)  # type: ignore

    return destination_metrics


def publish_without_correlation_id_header(destination: str, body: Dict, attempt=1, persistent=True):
    """
    Publishes a message without correlation-id on the headers.
    """
    publisher = build_publisher()

    standard_header = {
        "tshoot-destination": destination,
        # RabbitMQ
        # These two parameters must be set on consumer side as well, otherwise you'll get precondition_failed
        "x-dead-letter-routing-key": create_dlq_destination_from_another_destination(destination),
        "x-dead-letter-exchange": "",
    }

    if persistent:
        standard_header.update({"persistent": "true"})

    send_params = {
        "destination": destination,
        "body": json.dumps(body, cls=DjangoJSONEncoder),
        "headers": standard_header,
        "content_type": "application/json;charset=utf-8",
        "transaction": getattr(publisher, "_tmp_transaction_id", None),
    }
    send_params = clean_dict_with_falsy_or_strange_values(send_params)

    def _internal_send_logic():
        publisher.start_if_not_open()
        publisher.connection.send(**send_params)

    retry(_internal_send_logic, attempt=attempt)


def publish_to_destination(
    destination: str, body: Dict, headers: Dict = None, persistent: bool = True, attempts: int = 1
) -> None:
    """
    Publishes a message to a destination.
    """
    with build_publisher(f"random-publisher-{uuid4()}").auto_open_close_connection() as publisher:
        publisher.send(body=body, queue=destination, headers=headers, persistent=persistent, attempt=attempts)


def get_latest_message_from_destination_using_test_listener(destination: str) -> Dict:
    """
    Gets the latest message using the test listener utility. It does not ack the message
    on the destination queue.

    [!]: It makes a test hang forever if a message never arrives at the destination.
    """
    evaluation_consumer = build_listener(destination, is_testing=True)
    test_listener = evaluation_consumer._test_listener
    evaluation_consumer.start(wait_forever=False)

    # may wait forever if the msg never arrives
    test_listener.wait_for_message()
    received_message = test_listener.get_latest_message()

    return received_message


def iterable_len(iterable):
    """
    Calculates the length of any iterable (iterators included!)
    """
    return sum(1 for _ in iterable)


def wait_for_message_in_log(caplog, message_to_wait, message_count_to_wait=None, max_seconds_to_wait=None):
    """
    Awaits for a message that must appears for a given number of times.

    Args:
        caplog: An instance of LogCaptureFixture from pytest that is used to retrieve a list of
            format-interpolated log messages.
            Refer to: https://docs.pytest.org/en/latest/reference.html#_pytest.logging.LogCaptureFixture
        message_to_wait: A string message that'll be searched in the logs.
        message_count_to_wait: Optionally integer parameter that indicates how many `message_to_wait` we
            need to find in the logs. Defaults to 1.
        max_seconds_to_wait: Optionally integer parameter that indicates how many seconds the search will
            awaits for the messages appears in the logs. Defaults to 5 seconds
    """
    max_seconds_to_wait = max_seconds_to_wait or 5
    message_count_to_wait = message_count_to_wait or 1

    while max_seconds_to_wait:
        message_in_logs_count = iterable_len(
            filter(lambda message: re.compile(message_to_wait).match(message), caplog.messages)
        )
        if message_in_logs_count == message_count_to_wait:
            break
        max_seconds_to_wait -= 1
        sleep(1)


def get_active_threads_name_with_prefix(prefix: str):
    """
    Retrieves the threads name that starts with `prefix` and are active.
    """
    prefix_regex = re.compile(f"^{prefix}")
    return [thread.name for thread in threading.enumerate() if prefix_regex.match(thread.name)]


def is_testing_against_rabbitmq() -> bool:
    """
    Verify if the test suite is running against RabbitMQ.
    """
    try:
        rabbitmq.get_broker_version()
        return True
    except Exception:
        return False



from django.core.management.base import BaseCommand
from django_stomp.execution import start_processing


class Command(BaseCommand):
    help = "Start internal PUB/SUB logic"

    def add_arguments(self, parser):
        parser.add_argument("source_destination", type=str, help="Source destination used to consume messages")
        parser.add_argument(
            "callback_function",
            type=str,
            help="Entire module path with its function to process messages from source destination",
        )

    def handle(self, *args, **options):
        self.stdout.write(f"Provided parameters: {options}")

        source_destination = options.get("source_destination")
        callback_function = options.get("callback_function")

        self.stdout.write("Calling internal service to consume messages")
        start_processing(source_destination, callback_function)



import json
import logging
import ssl
import uuid
from contextlib import contextmanager
from typing import Dict
from typing import Optional

from django.core.serializers.json import DjangoJSONEncoder
from request_id_django_log.request_id import current_request_id
from request_id_django_log.settings import NO_REQUEST_ID
from stomp import Connection
from stomp.connect import StompConnection11

from django_stomp.helpers import clean_dict_with_falsy_or_strange_values
from django_stomp.helpers import create_dlq_destination_from_another_destination
from django_stomp.helpers import retry
from django_stomp.helpers import slow_down

logger = logging.getLogger("django_stomp")


class Publisher:
    """
    Class used to publish messages to brokers using the STOMP protocol. Some headers are removed
    if they are in the send() method as they cause unexpected behavior/errors.

    Such headers are defined in the UNSAFE_OR_RESERVED_BROKER_HEADERS_FOR_REMOVAL class variable which is used
    for sanitizing the user-supplied headers.
    """

    UNSAFE_OR_RESERVED_BROKER_HEADERS_FOR_REMOVAL = [
        # RabbitMQ unsafe headers
        "message-id",
        "transaction",
        "redelivered",
        "subscription",
        # Unsafe in a way that only django-stomp/broker should create
        "destination",
        "content-length",
        "content-type",
    ]

    def __init__(self, connection: StompConnection11, connection_configuration: Dict) -> None:
        self._connection_configuration = connection_configuration
        self.connection = connection
        self._default_content_type = "application/json;charset=utf-8"

    def is_open(self):
        return self.connection.is_connected()

    @slow_down
    def start(self):
        self.connection.connect(**self._connection_configuration)
        logger.info("Connected")

    def close(self):
        disconnect_receipt = str(uuid.uuid4())
        self.connection.disconnect(receipt=disconnect_receipt)
        logger.info("Disconnected")

    def start_if_not_open(self):
        if not self.is_open():
            logger.info("It is not open. Starting...")
            self.start()

    def send(self, body: dict, queue: str, headers=None, persistent=True, attempt=10):
        """
        Builds the final message headers/body and sends to the broker with the STOMP protocol. Attempt retries
        are ignored if the publisher is currently being used in a transaction in order to avoid already closed
        transactions errors due to the STOMP protocol behavior.
        """
        headers = self._build_final_headers(queue, headers, persistent)
        send_data = self._build_send_data(queue, body, headers)

        if self._is_publisher_in_transaction():
            self._send_to_broker_without_retry_attempts(send_data)
        else:
            self._send_to_broker(send_data, how_many_attempts=attempt)

    def _build_final_headers(self, queue: str, headers: Optional[Dict], persistent: bool) -> Dict:
        """
        Builds the message final headers. Removes unsafe or broker-reserved headers.
        Standard headers values override headers values to reduce possible errors.
        """
        if headers is None:
            headers = {}

        standard_headers = {
            "correlation-id": self._get_correlation_id(headers),
            "tshoot-destination": queue,
            # RabbitMQ
            # These two parameters must be set on consumer side as well, otherwise you'll get precondition_failed
            "x-dead-letter-routing-key": create_dlq_destination_from_another_destination(queue),
            "x-dead-letter-exchange": "",
        }

        # safety: standard_headers must override headers values, so order MATTERS here!
        mixed_headers = {**headers, **standard_headers}

        if persistent:
            self._add_persistent_messaging_header(mixed_headers)

        final_headers = self._remove_unsafe_or_reserved_for_broker_use_headers(mixed_headers)
        return final_headers

    def _get_correlation_id(self, headers: Optional[Dict]) -> str:
        """
        Gets the correlation id for the message. If 'correlation-id' is in the headers, this value is used.
        Otherwise, the value of current_request_id() is returned or a new one is generated as a last resort.
        """
        if headers and "correlation-id" in headers:
            return headers["correlation-id"]

        correlation_id = current_request_id() if current_request_id() != NO_REQUEST_ID else uuid.uuid4()
        return correlation_id

    def _remove_unsafe_or_reserved_for_broker_use_headers(self, headers: Dict) -> Dict:
        """
        Removes headers that are used internally by the brokers or that might cause unexpected behaviors (unsafe).
        """
        clean_headers = {
            key: headers[key] for key in headers if key not in self.UNSAFE_OR_RESERVED_BROKER_HEADERS_FOR_REMOVAL
        }

        return clean_headers

    def _build_send_data(self, queue: str, body: Dict, headers: Dict) -> Dict:
        """
        Builds the final data shape required to send messages using the STOMP protocol.
        """
        send_data = {
            "destination": queue,
            "body": json.dumps(body, cls=DjangoJSONEncoder),
            "headers": headers,
            "content_type": self._default_content_type,
            "transaction": getattr(self, "_tmp_transaction_id", None),
        }

        send_data = clean_dict_with_falsy_or_strange_values(send_data)
        return send_data

    def _send_to_broker(self, send_data: Dict, how_many_attempts: int) -> None:
        """
        Sends the actual data to the broker using the STOMP protocol with some retry attempts if
        a connection problem occurs.
        """

        def _internal_send_logic():
            self.start_if_not_open()
            self.connection.send(**send_data)

        retry(_internal_send_logic, attempt=how_many_attempts)

    def _send_to_broker_without_retry_attempts(self, send_data: Dict) -> None:
        """
        Sends the actual data to the broker using the STOMP protocol WITHOUT any retry attempts as reconnecting to the broker
        while a transaction was previously created will lead to 'bad transaction' errors because STOMP 1.1 protocol closes any
        transactions if the producer had TCP connection problems or sends a DISCONNECT frame.

        Hence, when a producer sends a BEGIN frame, all subsequent SEND frames (messages) must always use the SAME connection that
        was used to start the transaction.

        -> STOMP 1.1 specification: https://stomp.github.io/stomp-specification-1.1.html#BEGIN
        """
        self.connection.send(**send_data)  # bare sending without retries

    def _is_publisher_in_transaction(self) -> bool:
        """
        Checks if the publisher is currently being used in a transaction.
        """
        return hasattr(self, "_tmp_transaction_id")  # attribute set by do_inside_transaction contextmanager

    @staticmethod
    def _add_persistent_messaging_header(headers: Dict) -> Dict:
        value = {"persistent": "true"}

        if headers:
            headers.update(value)
            return headers

        return value

    @contextmanager
    def auto_open_close_connection(self):
        try:
            self.start_if_not_open()
            yield self
        finally:
            if self.is_open():
                self.close()

    @contextmanager
    def do_inside_transaction(self):
        try:
            self.start_if_not_open()
            transaction_id = self.connection.begin()
            logger.debug("Created transaction ID: %s", transaction_id)
            setattr(self, "_tmp_transaction_id", transaction_id)
            yield self
            self.connection.commit(transaction_id)
        except BaseException as e:
            logger.exception("Error inside transaction")
            if hasattr(self, "_tmp_transaction_id"):
                self.connection.abort(getattr(self, "_tmp_transaction_id"))
            raise e
        finally:
            if hasattr(self, "_tmp_transaction_id"):
                delattr(self, "_tmp_transaction_id")


def build_publisher(**connection_params) -> Publisher:
    logger.info("Building publisher...")
    hosts, vhost = [(connection_params.get("host"), connection_params.get("port"))], connection_params.get("vhost")
    if connection_params.get("hostStandby") and connection_params.get("portStandby"):
        hosts.append((connection_params.get("hostStandby"), connection_params.get("portStandby")))
    use_ssl = connection_params.get("use_ssl", False)
    ssl_version = connection_params.get("ssl_version", ssl.PROTOCOL_TLS)
    logger.info(f"Use SSL? {use_ssl}. Version: {ssl_version}")
    client_id = connection_params.get("client_id", uuid.uuid4())
    connection_configuration = {
        "username": connection_params.get("username"),
        "passcode": connection_params.get("password"),
        "wait": True,
        "headers": {"client-id": f"{client_id}-publisher"},
    }
    conn = Connection(hosts, ssl_version=ssl_version, use_ssl=use_ssl, vhost=vhost)
    publisher = Publisher(conn, connection_configuration)
    return publisher


@contextmanager
def auto_open_close_connection(publisher: Publisher):
    try:
        publisher.start()
        yield
    finally:
        if publisher.is_open():
            publisher.close()


@contextmanager
def do_inside_transaction(publisher: Publisher):
    try:
        publisher.start_if_not_open()
        transaction_id = publisher.connection.begin()
        logger.debug("Created transaction ID: %s", transaction_id)
        setattr(publisher, "_tmp_transaction_id", transaction_id)
        yield
        publisher.connection.commit(transaction_id)
    except BaseException as e:
        logger.exception("Error inside transaction")
        if hasattr(publisher, "_tmp_transaction_id"):
            publisher.connection.abort(getattr(publisher, "_tmp_transaction_id"))
        raise e
    finally:
        if hasattr(publisher, "_tmp_transaction_id"):
            delattr(publisher, "_tmp_transaction_id")



"""
Module with unit tests written to assert producer (publisher) scenarios.
"""
import json
from unittest import mock
from uuid import uuid4

from django.core.serializers.json import DjangoJSONEncoder
from django_stomp.builder import build_publisher


def test_should_build_final_headers_without_removing_unsafe_headers(mocker):
    # mocks
    current_request_id = uuid4()
    mocked_current_request_id = mocker.patch("django_stomp.services.producer.current_request_id")
    mocked_current_request_id.return_value = current_request_id

    publisher = build_publisher("test-publisher")
    persistent = True
    queue = "/queue/test"
    headers = {
        "eventType": "sentPurchaseReceipt",
        "another": "header",
        "x-dead-letter-exchange": "",
        "tshoot-destination": "/queue/wrrrrong-again",
        "x-dead-letter-routing-key": "DLQ.wrong-really-wrong...",  # queue name should define this
    }

    expected_final_headers = {
        "another": "header",
        "correlation-id": current_request_id,
        "eventType": "sentPurchaseReceipt",
        "persistent": "true",
        "tshoot-destination": "/queue/test",
        "x-dead-letter-exchange": "",
        "x-dead-letter-routing-key": "DLQ.test",
    }

    final_headers = publisher._build_final_headers(queue, headers, persistent)

    assert final_headers == expected_final_headers


def test_should_build_final_headers_with_removal_of_unsafe_headers(mocker):
    # mocks
    current_request_id = uuid4()
    mocked_current_request_id = mocker.patch("django_stomp.services.producer.current_request_id")
    mocked_current_request_id.return_value = current_request_id

    publisher = build_publisher("test-publisher")
    persistent = True
    queue = "/queue/some-destination"

    # no correlation-id supplied here
    raw_headers = {
        "subscription": "77625173-42b5-4c62-9c24-1fc4a4668dcc-listener",
        "destination": "/queue/some-destination",
        "message-id": "T_77625173-42b5-4c62-9c24-1fc4a4668dcc-listener@@session-jSfP4eyH59eKxHaZDv59Cg@@2",
        "redelivered": "false",
        "x-dead-letter-routing-key": "DLQ.wrong-really-wrong-destination",
        "x-dead-letter-exchange": "",
        "tshoot-destination": "/queue/wrong-really-wrong-destination",
        "transaction": "11837f0b-f1c5-494f-9b2b-347c636d8ca2",
        "eventType": "someType",
        "persistent": "true",
        "content-type": "application/json;charset=utf-8",
        "content-length": "298",
    }

    expected_final_headers = {
        "persistent": "true",
        "tshoot-destination": "/queue/some-destination",
        "x-dead-letter-routing-key": "DLQ.some-destination",
        "x-dead-letter-exchange": "",
        "eventType": "someType",
        "correlation-id": current_request_id,
    }

    final_headers = publisher._build_final_headers(queue, raw_headers, persistent)

    assert final_headers == expected_final_headers


def test_should_remove_unsafe_headers():
    publisher = build_publisher("test-publisher")

    unsafe_headers = {
        "subscription": "77625173-42b5-4c62-9c24-1fc4a4668dcc-listener",
        "destination": "/queue/some-destination",
        "message-id": "T_77625173-42b5-4c62-9c24-1fc4a4668dcc-listener@@session-jSfP4eyH59eKxHaZDv59Cg@@2",
        "redelivered": "false",
        "x-dead-letter-routing-key": "DLQ.some-destination",
        "x-dead-letter-exchange": "",
        "tshoot-destination": "/queue/some-destination",
        "transaction": "11837f0b-f1c5-494f-9b2b-347c636d8ca2",
        "eventType": "someType",
        "correlation-id": "6e10903e13ddcde54304883e5df713a5",
        "persistent": "true",
        "content-type": "application/json;charset=utf-8",
        "content-length": "298",
    }

    final_headers = {
        "persistent": "true",
        "tshoot-destination": "/queue/some-destination",
        "x-dead-letter-routing-key": "DLQ.some-destination",
        "x-dead-letter-exchange": "",
        "eventType": "someType",
        "correlation-id": "6e10903e13ddcde54304883e5df713a5",
    }

    # this test already considers a safe tshoot-destination, x-dead-letter-routing-key as
    # this internal method is invoked after standard_headers (safe) are imposed to the unsafe_headers
    assert publisher._remove_unsafe_or_reserved_for_broker_use_headers(unsafe_headers) == final_headers


def test_should_build_send_data_for_broker():
    publisher = build_publisher("test-publisher")

    clean_headers = {"eventType": "goal"}
    queue = "/queue/test"
    body = {"points": "100.0000"}

    expected_send_data = {
        "destination": queue,
        "body": json.dumps(body, cls=DjangoJSONEncoder),
        "headers": {"eventType": "goal"},
        "content_type": "application/json;charset=utf-8",
    }

    send_data = publisher._build_send_data(queue, body, clean_headers)
    assert expected_send_data == send_data



import json
import logging
import ssl
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from enum import Enum
from typing import Callable
from typing import Dict
from typing import Optional

import stomp
from stomp import connect
from stomp.utils import Frame as StompFrame

from django_stomp.helpers import create_dlq_destination_from_another_destination
from django_stomp.helpers import is_heartbeat_enabled
from django_stomp.helpers import only_destination_name
from django_stomp.settings import STOMP_PROCESS_MSG_WORKERS

logger = logging.getLogger("django_stomp")


class Acknowledgements(Enum):
    """
    See more details at:
        - https://pubsub.github.io/pubsub-specification-1.2.html#SUBSCRIBE_ack_Header
        - https://jasonrbriggs.github.io/pubsub.py/api.html#acks-and-nacks
    """

    CLIENT = "client"
    CLIENT_INDIVIDUAL = "client-individual"
    AUTO = "auto"


@dataclass(frozen=True)
class Payload:
    ack: Callable
    nack: Callable
    headers: Dict
    body: Dict


class Listener(stomp.ConnectionListener):
    def __init__(
        self,
        connection: connect.StompConnection11,
        callback: Callable,
        subscription_configuration: Dict,
        connection_configuration: Dict,
        should_process_msg_on_background: bool,
        is_testing: bool = False,
        subscription_id: Optional[str] = None,
    ) -> None:
        self._subscription_configuration = subscription_configuration
        self._connection_configuration = connection_configuration
        self._connection = connection
        self._callback = callback
        self._subscription_id = f"{subscription_id if subscription_id else str(uuid.uuid4())}-listener"
        self._listener_id = str(uuid.uuid4())
        self._is_testing = is_testing
        self._should_process_msg_on_background = should_process_msg_on_background
        self._pool_executor = self._create_new_worker_executor()

        if self._is_testing:
            from stomp.listener import TestListener

            self._test_listener = TestListener()
        else:
            self._test_listener = None

    def _create_new_worker_executor(self):
        return ThreadPoolExecutor(max_workers=STOMP_PROCESS_MSG_WORKERS, thread_name_prefix=self._subscription_id)

    def on_message(self, frame: StompFrame):
        headers, body = frame.headers, frame.body

        message_id = headers["message-id"]
        logger.info(f"Message ID: {message_id}")
        logger.debug("Received headers: %s", headers)
        logger.debug("Received message: %s", body)

        # https://jasonrbriggs.github.io/stomp.py/api.html#acks-and-nacks
        def ack_logic():
            self._connection.ack(message_id, self._subscription_id)

        def nack_logic():
            # Requeue is used because of RabbitMQ: https://www.rabbitmq.com/stomp.html#ack-nack
            self._connection.nack(message_id, self._subscription_id, requeue=False)

        payload = Payload(ack_logic, nack_logic, headers, json.loads(body))

        if self._should_process_msg_on_background:
            self._submit_task_to_worker_pool(payload)
        else:
            self._callback(payload)

    def _submit_task_to_worker_pool(self, payload):
        try:
            self._pool_executor.submit(self._callback, payload)
        except RuntimeError:
            logger.warning("Worker pool was shutdown!")
            self._pool_executor = self._create_new_worker_executor()
            self._pool_executor.submit(self._callback, payload)

    def is_open(self):
        return self._connection.is_connected()

    def start(self, callback: Callable = None, wait_forever=True):
        logger.info(f"Starting listener with name: {self._listener_id}")
        logger.info(f"Subscribe/Listener auto-generated ID: {self._subscription_id}")

        self._set_listener()

        self._callback = callback if callback else self._callback
        self._connection.connect(**self._connection_configuration)
        self._connection.subscribe(
            id=self._subscription_id,
            headers=self._connection_configuration["headers"],
            **self._subscription_configuration,
        )
        logger.info("Connected")
        if wait_forever:
            while True:
                if not self.is_open():
                    logger.info("It is not open. Starting...")
                    self.start(self._callback, wait_forever=False)
                time.sleep(1)

    def _set_listener(self):
        if self._is_testing:
            self._connection.set_listener("TESTING", self._test_listener)
        else:
            self._connection.set_listener(self._listener_id, self)

    def close(self):
        disconnect_receipt = str(uuid.uuid4())
        self._connection.disconnect(receipt=disconnect_receipt)
        logger.info("Disconnected")

    def shutdown_worker_pool(self):
        self._pool_executor.shutdown()

    def on_error(self, frame: StompFrame):
        logger.warning("Received ERROR frame from broker - headers: %d - body: %d", frame.headers, frame.body)


def build_listener(
    destination_name,
    should_process_msg_on_background,
    callback=None,
    ack_type=Acknowledgements.CLIENT,
    durable_topic_subscription=False,
    is_testing=False,
    routing_key=None,
    **connection_params,
) -> Listener:
    logger.info("Building listener...")
    hosts, vhost = [(connection_params.get("host"), connection_params.get("port"))], connection_params.get("vhost")
    if connection_params.get("hostStandby") and connection_params.get("portStandby"):
        hosts.append((connection_params.get("hostStandby"), connection_params.get("portStandby")))
    use_ssl = connection_params.get("use_ssl", False)
    ssl_version = connection_params.get("ssl_version", ssl.PROTOCOL_TLS)
    outgoing_heartbeat = int(connection_params.get("outgoingHeartbeat", 0))
    incoming_heartbeat = int(connection_params.get("incomingHeartbeat", 0))

    logger.info(
        f"Use SSL? {use_ssl}. Version: {ssl_version}. Outgoing/Ingoing heartbeat: {outgoing_heartbeat}/{incoming_heartbeat}. Background? {should_process_msg_on_background}"
    )

    if is_heartbeat_enabled(outgoing_heartbeat, incoming_heartbeat) and not should_process_msg_on_background:
        logger.warning(
            "STOMP heartbeat enabled while message processing on background is disable! "
            "This could potentially lead to a false positive heartbeat timeout!"
        )

    # http://stomp.github.io/stomp-specification-1.2.html#Heart-beating
    # http://jasonrbriggs.github.io/stomp.py/api.html
    conn = connect.StompConnection11(
        hosts,
        ssl_version=ssl_version,
        use_ssl=use_ssl,
        heartbeats=(outgoing_heartbeat, incoming_heartbeat),
        vhost=vhost,
    )
    client_id = connection_params.get("client_id", uuid.uuid4())
    routing_key = routing_key or destination_name
    subscription_configuration = {
        "destination": routing_key,
        "ack": ack_type.value,
        # RabbitMQ
        "x-queue-name": only_destination_name(destination_name),
        "auto-delete": "false",
        "durable": "true",
    }
    header_setup = {
        # ActiveMQ
        "client-id": f"{client_id}-listener",
        "activemq.prefetchSize": "1",
        # RabbitMQ
        "prefetch-count": "1",
        # These two parameters must be set on producer side as well, otherwise you'll get precondition_failed
        "x-dead-letter-routing-key": create_dlq_destination_from_another_destination(destination_name),
        "x-dead-letter-exchange": "",
    }

    if durable_topic_subscription is True:
        durable_subs_header = {
            # ActiveMQ
            "activemq.subscriptionName": header_setup["client-id"],
            "activemq.subcriptionName": header_setup["client-id"],
        }
        header_setup.update(durable_subs_header)
    connection_configuration = {
        "username": connection_params.get("username"),
        "passcode": connection_params.get("password"),
        "wait": True,
        "headers": header_setup,
    }
    listener = Listener(
        conn,
        callback,
        subscription_configuration,
        connection_configuration,
        is_testing=is_testing,
        subscription_id=connection_params.get("subscriptionId"),
        should_process_msg_on_background=should_process_msg_on_background,
    )
    return listener



from typing import Any
from typing import Callable
from typing import Optional

from django.conf import settings as django_settings
from django_stomp.exceptions import DjangoStompImproperlyConfigured
from django_stomp.helpers import eval_as_int_if_provided_value_is_not_none_otherwise_none


def eval_settings_otherwise_raise_exception(
    settings_name: str, evaluation_callback: Callable, default_value: Optional[Any] = None
):
    try:
        return evaluation_callback(getattr(django_settings, settings_name, default_value))
    except Exception:
        raise DjangoStompImproperlyConfigured(f"The defined {settings_name} is not valid!")


STOMP_PROCESS_MSG_WORKERS = eval_settings_otherwise_raise_exception(
    "STOMP_PROCESS_MSG_WORKERS", eval_as_int_if_provided_value_is_not_none_otherwise_none
)



from io import StringIO

import pytest
from django.core.management import CommandError
from django.core.management import call_command
from pytest_mock import MockFixture


@pytest.fixture
def mock_send_message_from_one_destination_to_another(mocker: MockFixture):
    return mocker.patch("django_stomp.management.commands.move_messages.send_message_from_one_destination_to_another")


def test_should_call_logic_to_send_one_destination_to_another(mock_send_message_from_one_destination_to_another):
    out = StringIO()
    fake_source = "/queue/your-source"
    fake_target = "/queue/your-target"

    call_command("move_messages", fake_source, fake_target, stdout=out)

    assert "Calling internal service to send messages to target" in out.getvalue()
    mock_send_message_from_one_destination_to_another.assert_called_with(
        fake_source, fake_target, custom_stomp_server_host=None, custom_stomp_server_port=None
    )


def test_should_raise_error_if_essential_parameters_are_missing(mock_send_message_from_one_destination_to_another):
    fake_source = "/queue/your-source"

    with pytest.raises(CommandError) as e:
        call_command("move_messages")

    assert e.value.args[0] == "Error: the following arguments are required: source_destination, target_destination"

    with pytest.raises(CommandError) as e:
        call_command("move_messages", fake_source)

    assert e.value.args[0] == "Error: the following arguments are required: target_destination"

    mock_send_message_from_one_destination_to_another.assert_not_called()


def test_should_call_logic_to_send_one_destination_to_another_given_custom_broker(
    mock_send_message_from_one_destination_to_another,
):
    out = StringIO()
    fake_source = "/queue/your-source"
    fake_target = "/queue/your-target"
    fake_custom_broker_host = "jafar-host"
    fake_custom_broker_port = "61614"

    call_command(
        "move_messages", fake_source, fake_target, fake_custom_broker_host, fake_custom_broker_port, stdout=out
    )

    assert "Calling internal service to send messages to target" in out.getvalue()
    mock_send_message_from_one_destination_to_another.assert_called_with(
        fake_source,
        fake_target,
        custom_stomp_server_host=fake_custom_broker_host,
        custom_stomp_server_port=int(fake_custom_broker_port),
    )


def test_should_raise_error_given_custom_port_is_not_informed_to_use_custom_broker(
    mock_send_message_from_one_destination_to_another,
):
    out = StringIO()
    fake_source = "/queue/your-source"
    fake_target = "/queue/your-target"
    fake_custom_broker_host = "jafar-host"

    with pytest.raises(CommandError) as e:
        call_command("move_messages", fake_source, fake_target, fake_custom_broker_host, stdout=out)

    assert e.value.args[0] == "You should provide broker_port_for_source_destination option as well"

    mock_send_message_from_one_destination_to_another.assert_not_called()



import logging
from typing import Dict
from typing import Optional

from django.conf import settings
from django_stomp.helpers import clean_dict_with_falsy_or_strange_values
from django_stomp.helpers import eval_as_int_otherwise_none
from django_stomp.helpers import eval_str_as_boolean
from django_stomp.services import consumer
from django_stomp.services import producer
from django_stomp.services.consumer import Listener
from django_stomp.services.producer import Publisher

logger = logging.getLogger("django_stomp")


def build_publisher(client_id: Optional[str] = None) -> Publisher:
    connection_params = _build_connection_parameter(client_id)

    return producer.build_publisher(**connection_params)


def build_listener(
    destination_name: str,
    durable_topic_subscription: bool = False,
    should_process_msg_on_background: bool = True,
    is_testing: bool = False,
    client_id: Optional[str] = None,
    routing_key: Optional[str] = None,
    custom_stomp_server_host: Optional[str] = None,
    custom_stomp_server_port: Optional[int] = None,
) -> Listener:
    connection_params = _build_connection_parameter(client_id, custom_stomp_server_host, custom_stomp_server_port)

    return consumer.build_listener(
        destination_name,
        durable_topic_subscription=durable_topic_subscription,
        is_testing=is_testing,
        routing_key=routing_key,
        should_process_msg_on_background=should_process_msg_on_background,
        **connection_params,
    )


def _build_connection_parameter(
    client_id: Optional[str] = None,
    custom_stomp_server_host: Optional[str] = None,
    custom_stomp_server_port: Optional[int] = None,
) -> Dict:
    stomp_server_host = custom_stomp_server_host or getattr(settings, "STOMP_SERVER_HOST", None)
    stomp_server_port = eval_as_int_otherwise_none(custom_stomp_server_port) or eval_as_int_otherwise_none(
        getattr(settings, "STOMP_SERVER_PORT", None)
    )
    stomp_server_standby_port = eval_as_int_otherwise_none(getattr(settings, "STOMP_SERVER_STANDBY_PORT", None))
    outgoing_heartbeat = eval_as_int_otherwise_none(getattr(settings, "STOMP_OUTGOING_HEARTBEAT", "10000"))
    incoming_heartbeat = eval_as_int_otherwise_none(getattr(settings, "STOMP_INCOMING_HEARTBEAT", "10000"))
    subscription_id = getattr(settings, "STOMP_SUBSCRIPTION_ID", None)

    required_params = {
        "host": stomp_server_host,
        "port": stomp_server_port,
        "hostStandby": getattr(settings, "STOMP_SERVER_STANDBY_HOST", None),
        "portStandby": stomp_server_standby_port,
        "outgoingHeartbeat": outgoing_heartbeat,
        "incomingHeartbeat": incoming_heartbeat,
        "subscriptionId": subscription_id,
        "vhost": getattr(settings, "STOMP_SERVER_VHOST", None),
    }
    extra_params = {"use_ssl": eval_str_as_boolean(settings.STOMP_USE_SSL), "client_id": client_id}

    logger.info("Server details connection: %s. Extra params: %s", required_params, extra_params)

    credentials = {
        "username": getattr(settings, "STOMP_SERVER_USER", None),
        "password": getattr(settings, "STOMP_SERVER_PASSWORD", None),
    }

    return clean_dict_with_falsy_or_strange_values({**required_params, **extra_params, **credentials})



from time import sleep

import requests
from parsel import Selector
from tests.support.dtos import CurrentDestinationStatus


def current_queue_configuration(queue_name, host="localhost") -> CurrentDestinationStatus:
    sleep(1)
    result = requests.get(f"http://{host}:8161/admin/queues.jsp", auth=("admin", "admin"))
    selector = Selector(text=str(result.content))

    all_queues = selector.xpath('//*[@id="queues"]/tbody/tr').getall()

    assert len(all_queues) > 0
    for index, queue_details in enumerate(all_queues):
        queue_details_as_selector = Selector(text=queue_details)
        if f"JMSDestination={queue_name}" in queue_details_as_selector.css("td a::attr(href)").get():
            n_of_pending_messages = int(queue_details_as_selector.css("td + td::text").get())
            n_of_consumers = int(queue_details_as_selector.css("td + td + td::text").get())
            m_enqueued = int(queue_details_as_selector.css("td + td + td + td::text").get())
            m_dequeued = int(queue_details_as_selector.css("td + td + td + td + td::text").get())
            return CurrentDestinationStatus(n_of_pending_messages, n_of_consumers, m_enqueued, m_dequeued)
        if all_queues[index] == all_queues[-1]:
            raise Exception



import functools
import logging
import time
import uuid
from typing import Callable
from typing import Dict

import tenacity

logger = logging.getLogger("django_stomp")


def slow_down(_func=None, *args, **kwargs):
    """Sleep given amount of seconds before calling the function"""

    before = kwargs.get("before", 0.5)
    after = kwargs.get("after", 0.5)

    def decorator_slow_down(func):
        @functools.wraps(func)
        def wrapper_slow_down(*args, **kwargs):
            time.sleep(before)
            value = func(*args, **kwargs)
            time.sleep(after)
            return value

        return wrapper_slow_down

    if _func is None:
        return decorator_slow_down
    else:
        return decorator_slow_down(_func)


def eval_str_as_boolean(value: str):
    return str(value).lower() in ("true", "1", "t", "y")


def return_none_if_provided_value_is_falsy_or_strange(value):
    if value is not None and (value == "" or value == b"" or value in (".", "none")):
        return None
    return value


def clean_dict_with_falsy_or_strange_values(value: Dict) -> Dict:
    return {k: v for k, v in value.items() if return_none_if_provided_value_is_falsy_or_strange(v) or v == 0}


def eval_as_int_otherwise_none(value):
    return int(value) if value or value == 0 else None


def only_destination_name(destination: str) -> str:
    position = destination.rfind("/")
    if position > 0:
        return destination[position + 1 :]
    return destination


def create_dlq_destination_from_another_destination(destination: str) -> str:
    return f"DLQ.{only_destination_name(destination)}"


def remove_key_from_dict(dictionary, key):
    dictionary.pop(key, None)


def is_destination_from_virtual_topic(destination_name: str) -> bool:
    return ".VirtualTopic." in destination_name


def is_dlq_destination(destination_name: str) -> bool:
    return "DLQ." in destination_name


def get_subscription_destination(destination_name: str) -> str:
    """
    Given a destination name like Consumer.XPTO.VirtualTopic.topic-name, returns
    '/topic/VirtualTopic.topic-name' in order to mimic the ActiveMQ Virtual Topics
    default behaviour in RabbitMQ.

    More on: https://activemq.apache.org/virtual-destinations
    """
    if is_destination_from_virtual_topic(destination_name) and not is_dlq_destination(destination_name):
        virtual_topic_name = destination_name.split(".VirtualTopic.")[-1]
        return f"/topic/VirtualTopic.{virtual_topic_name}"
    return destination_name


def get_listener_client_id(durable_topic_subscription: bool, listener_client_id: str) -> str:
    if not durable_topic_subscription and listener_client_id:
        return f"{listener_client_id}-{uuid.uuid4().hex}"
    return listener_client_id


def retry(function: Callable, attempt=10, *args, **kwargs):
    retry_configuration = tenacity.Retrying(
        stop=tenacity.stop_after_attempt(attempt),
        wait=tenacity.wait_fixed(3) + tenacity.wait_random(0, 2),
        after=tenacity.after_log(logger, logging.WARNING) if logger else None,
        reraise=True,
    )
    return retry_configuration(function, *args, **kwargs)


def eval_as_int_if_provided_value_is_not_none_otherwise_none(value):
    return int(value) if value is not None else None


def is_heartbeat_enabled(outgoing_heartbeat: int, incoming_heartbeat: int):
    """
    Determine if STOMP heartbeat is enabled or not. Per the specification, it'll only be enabled
    if a both estabilished times is greater than zero.

    More on: https://stomp.github.io/stomp-specification-1.1.html#Heart-beating
    """
    return outgoing_heartbeat > 0 and incoming_heartbeat > 0



from time import sleep

import requests

from parsel import Selector
from tests.support.dtos import CurrentDestinationStatus


def current_topic_configuration(topic_name, host="localhost") -> CurrentDestinationStatus:
    sleep(1)
    result = requests.get(f"http://{host}:8161/admin/topics.jsp", auth=("admin", "admin"))
    selector = Selector(text=str(result.content))

    all_topics = selector.xpath('//*[@id="topics"]/tbody/tr').getall()

    assert len(all_topics) > 0
    for index, queue_details in enumerate(all_topics):
        topic_details_as_selector = Selector(text=queue_details)
        if f"JMSDestination={topic_name}" in topic_details_as_selector.css("td a::attr(href)").get():
            number_of_consumers = int(topic_details_as_selector.css("td + td::text").get())
            messages_enqueued = int(topic_details_as_selector.css("td + td + td::text").get())
            messages_dequeued = int(topic_details_as_selector.css("td + td + td + td::text").get())
            return CurrentDestinationStatus(None, number_of_consumers, messages_enqueued, messages_dequeued)
        if all_topics[index] == all_topics[-1]:
            raise Exception



from dataclasses import dataclass
from typing import Dict
from typing import Optional


@dataclass(frozen=True)
class CurrentDestinationStatus:
    number_of_pending_messages: Optional[int]
    number_of_consumers: int
    messages_enqueued: int
    messages_dequeued: int


@dataclass(frozen=True)
class ConsumerStatus:
    address_to_destination_details: Optional[str]
    destination_name: str
    session_id: Optional[int]
    enqueues: Optional[int]
    dequeues: Optional[int]
    dispatched: Optional[int]
    dispatched_queue: Optional[int]
    prefetch: int
    max_pending: Optional[int]
    exclusive: bool
    retroactive: Optional[bool]


@dataclass(frozen=True)
class MessageStatus:
    message_id: Optional[str]
    details: Dict
    persistent: Optional[bool]
    correlation_id: str
    properties: Optional[Dict]


@dataclass(frozen=True)
class SubscriberSetup:
    address_to_subscriber_details: str
    subscriber_id: str
    destination: str
    pending_queue_size: int
    dispatched_queue_size: int
    dispatched_counter: int
    enqueue_counter: int
    dequeue_counter: int



import os

from setuptools import find_packages
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="django-stomp",
    version="5.0.0",
    description="A simple implementation of STOMP with Django",
    long_description=README,
    long_description_content_type="text/markdown",
    include_package_data=True,
    author="Ricardo Baltazar Chaves, Willian Antunes",
    author_email="Ricardo Baltazar <ricardobchaves6@gmail.com>, Willian Antunes <willian.lima.antunes@gmail.com>",
    license="MIT",
    url="https://github.com/juntossomosmais/django-stomp",
    packages=find_packages(),
    install_requires=["request-id-django-log==0.1.1", "stomp.py~=7.0", "tenacity~=8.0"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Framework :: Django :: 2.1",
        "Environment :: Web Environment",
        "Natural Language :: Portuguese (Brazilian)",
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import json
import logging
import re
import threading
import uuid
from time import sleep
from unittest import mock
from uuid import uuid4

import pytest
import trio
from django.db.backends.signals import connection_created
from pytest_mock import MockFixture
from stomp.exception import NotConnectedException

from django_stomp.builder import build_listener
from django_stomp.builder import build_publisher
from django_stomp.execution import clean_messages_on_destination_by_acking
from django_stomp.execution import send_message_from_one_destination_to_another
from django_stomp.execution import start_processing
from django_stomp.services.producer import do_inside_transaction
from tests.support import rabbitmq
from tests.support.activemq.connections_details import consumers_details
from tests.support.activemq.message_details import retrieve_message_published
from tests.support.activemq.queue_details import current_queue_configuration
from tests.support.activemq.subscribers_details import offline_durable_subscribers
from tests.support.activemq.topic_details import current_topic_configuration
from tests.support.callbacks_for_tests import callback_move_and_ack_path
from tests.support.callbacks_for_tests import callback_standard_path
from tests.support.callbacks_for_tests import callback_with_another_log_message_path
from tests.support.callbacks_for_tests import callback_with_exception_path
from tests.support.callbacks_for_tests import callback_with_explicit_db_connection_path
from tests.support.callbacks_for_tests import callback_with_logging_path
from tests.support.callbacks_for_tests import callback_with_nack_path
from tests.support.callbacks_for_tests import callback_with_sleep_three_seconds_path
from tests.support.callbacks_for_tests import callback_with_sleep_three_seconds_while_heartbeat_thread_is_alive_path
from tests.support.helpers import get_destination_metrics_from_broker
from tests.support.helpers import get_latest_message_from_destination_using_test_listener
from tests.support.helpers import is_testing_against_rabbitmq
from tests.support.helpers import publish_to_destination
from tests.support.helpers import publish_without_correlation_id_header
from tests.support.helpers import wait_for_message_in_log


def test_should_consume_message_and_publish_to_another_queue_using_same_correlation_id():
    # Base environment setup
    destination_one = f"/queue/my-test-destination-one-{uuid4()}"
    destination_two = f"/queue/my-test-destination-two-{uuid4()}"
    destination_three = f"/queue/my-test-destination-three-{uuid4()}"

    some_correlation_id = uuid.uuid4()
    some_header = {"correlation-id": some_correlation_id}
    some_body = {"keyOne": 1, "keyTwo": 2}

    # publishes to destination one
    publish_to_destination(destination_one, some_body, some_header)

    # return_listener=True is required to avoid testing_listener.close() on the test's main thread which prematurely closes
    # the connection to the broker which can compromise the callback's ACK (runs on another thread)
    # consumes from destination_one and publishes to destination_two
    start_processing(
        destination_one,
        callback_move_and_ack_path,
        param_to_callback=destination_two,
        is_testing=True,
        return_listener=True,
    )

    # consumes from destination_two and publishes to destination_three
    start_processing(
        destination_two,
        callback_move_and_ack_path,
        param_to_callback=destination_three,
        is_testing=True,
        return_listener=True,
    )

    # asserts that the msg is received on destination three WITH THE SAME correlation-id of destination_one
    received_message = get_latest_message_from_destination_using_test_listener(destination_three)

    assert received_message is not None
    received_header = received_message[0]

    assert received_header["correlation-id"] == str(some_correlation_id)
    received_body = json.loads(received_message[1])

    assert received_body == some_body


def test_should_consume_message_and_publish_to_another_queue_using_creating_correlation_id(mocker: MockFixture):
    # Base environment setup
    some_body = {"keyOne": 1, "keyTwo": 2}

    destination_three = f"/queue/my-test-destination-again-three-{uuid4()}"
    destination_four = f"/queue/my-test-destination-again-four-{uuid4()}"

    # mocks: It must be called only once to generate the correlation-id
    mock_uuid = mocker.patch("django_stomp.services.producer.uuid")

    uuid_to_publisher, uuid_to_correlation_id, uuid_to_another_publisher, uuid_publisher_mover = (
        uuid.uuid4(),
        uuid.uuid4(),
        uuid.uuid4(),
        uuid.uuid4(),
    )

    mock_uuid.uuid4.side_effect = [
        uuid_to_publisher,
        uuid_to_correlation_id,
        uuid_to_another_publisher,
        uuid_publisher_mover,
    ]

    # publishes to destination three
    publish_to_destination(destination_three, some_body)

    # consumes from destination three and publishes to destination four
    start_processing(
        destination_three,
        callback_move_and_ack_path,
        param_to_callback=destination_four,
        is_testing=True,
        return_listener=True,
    )

    # asserts that the msg is received on destination four WITH THE SAME correlation-id of destination_three
    received_message = get_latest_message_from_destination_using_test_listener(destination_four)

    assert received_message is not None
    received_header = received_message[0]

    assert received_header["correlation-id"] == str(uuid_to_correlation_id)
    received_body = json.loads(received_message[1])

    assert received_body == some_body


def test_should_consume_message_and_dequeue_it_using_ack():
    # In order to publish sample data
    some_body = {"keyOne": 1, "keyTwo": 2}
    test_destination_consumer_one = f"/queue/my-test-destination-consumer-one-{uuid4()}"
    *_, queue_name = test_destination_consumer_one.split("/")

    publish_to_destination(test_destination_consumer_one, some_body)
    start_processing(test_destination_consumer_one, callback_standard_path, is_testing=True)
    queue_status = get_destination_metrics_from_broker(queue_name)

    assert queue_status.number_of_pending_messages == 0
    assert queue_status.number_of_consumers == 0
    assert queue_status.messages_enqueued == 1
    assert queue_status.messages_dequeued == 1


def test_should_create_durable_subscriber_and_receive_standby_messages(mocker: MockFixture, settings):
    test_destination_durable_consumer_one = f"/topic/my-test-destination-durable-consumer-one-{uuid4()}"
    temp_uuid_listener = str(uuid.uuid4())

    mocker.patch("django_stomp.execution.get_listener_client_id", return_value=temp_uuid_listener)
    mocker.patch("django_stomp.execution.durable_topic_subscription", True)
    durable_subscription_id = str(uuid.uuid4())
    settings.STOMP_SUBSCRIPTION_ID = durable_subscription_id
    # Just to create a durable subscription
    start_processing(test_destination_durable_consumer_one, callback_standard_path, is_testing=True)
    settings.STOMP_SUBSCRIPTION_ID = None

    # In order to publish sample data
    some_body = {"keyOne": 1, "keyTwo": 2}

    publish_to_destination(test_destination_durable_consumer_one, some_body)
    publish_to_destination(test_destination_durable_consumer_one, some_body)
    publish_to_destination(test_destination_durable_consumer_one, some_body)

    # To recreate a durable subscription
    settings.STOMP_SUBSCRIPTION_ID = durable_subscription_id
    start_processing(test_destination_durable_consumer_one, callback_standard_path, is_testing=True)

    *_, topic_name = test_destination_durable_consumer_one.split("/")
    try:
        destination_status = current_topic_configuration(topic_name)
        assert destination_status.number_of_consumers == 1
        assert destination_status.messages_enqueued == 3
        assert destination_status.messages_dequeued == 3

        all_offline_subscribers = list(offline_durable_subscribers("localhost"))
        assert len(all_offline_subscribers) > 0
        for index, subscriber_setup in enumerate(all_offline_subscribers):
            if subscriber_setup.subscriber_id == f"{temp_uuid_listener}-listener":
                assert subscriber_setup.dispatched_counter == 3
                assert subscriber_setup.enqueue_counter == 3
                assert subscriber_setup.dequeue_counter == 3
                break
            assert all_offline_subscribers[index] != all_offline_subscribers[-1]
    except Exception:
        destination_status = rabbitmq.current_topic_configuration(topic_name)  # type: ignore
        assert destination_status.number_of_consumers == 0
        assert destination_status.messages_enqueued == 3
        assert destination_status.messages_dequeued == 3


async def test_should_configure_prefetch_size_as_one(mocker: MockFixture, settings):
    """
    See more details here: https://activemq.apache.org/stomp.html

    Specifies the maximum number of pending messages that will be dispatched to the client.
    Once this maximum is reached no more messages are dispatched until the client acknowledges a
    message. Set to a low value > 1 for fair distribution of messages across consumers
    when processing messages can be slow. Note: if your STOMP client is implemented using a dynamic
    scripting language like Ruby, say, then this parameter must be set to 1 as there is no notion of a
    client-side message size to be sized. STOMP does not support a value of 0.
    """
    test_destination_prefetch_consumer_one = f"/queue/my-destination-prefetch-consumer-one-{uuid4()}"

    listener_id = str(uuid.uuid4())
    # LISTENER_CLIENT_ID is used by ActiveMQ
    mocker.patch("django_stomp.execution.get_listener_client_id", return_value=listener_id)
    # SUBSCRIPTION_ID is used by RabbitMQ to identify a consumer through a tag
    subscription_id = str(uuid.uuid4())
    settings.STOMP_SUBSCRIPTION_ID = subscription_id

    async def collect_consumer_details():
        await trio.sleep(2)
        try:
            consumers = list(consumers_details(f"{listener_id}-listener"))
        except Exception:
            consumers = list(rabbitmq.consumers_details(f"{subscription_id}-listener"))

        assert len(consumers) > 0
        for index, consumer_status in enumerate(consumers):
            if consumer_status.destination_name in test_destination_prefetch_consumer_one:
                assert consumer_status.prefetch == 1
                assert consumer_status.max_pending == 0
                break
            assert consumers[index] != consumers[-1]

    async def execute_start_processing():
        listener = start_processing(
            test_destination_prefetch_consumer_one,
            callback_standard_path,
            is_testing=True,
            testing_disconnect=False,
            return_listener=True,
        )
        await trio.sleep(10)
        listener.close()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(execute_start_processing)
        nursery.start_soon(collect_consumer_details)


def test_should_publish_to_dql_due_to_explicit_nack():
    test_destination_dlq_one = f"/queue/my-destination-dql-one-{uuid.uuid4()}"
    *_, queue_name = test_destination_dlq_one.split("/")
    dlq_queue_name = f"DLQ.{queue_name}"

    # In order to publish sample data
    some_body = {"keyOne": 1, "keyTwo": 2}
    publish_to_destination(test_destination_dlq_one, some_body)

    start_processing(test_destination_dlq_one, callback_with_nack_path, is_testing=True)

    try:
        queue_status = current_queue_configuration(dlq_queue_name)

        assert queue_status.number_of_pending_messages == 1
        assert queue_status.number_of_consumers == 0
        assert queue_status.messages_enqueued == 1
        assert queue_status.messages_dequeued == 0
    except Exception:
        queue_status = rabbitmq.current_queue_configuration(dlq_queue_name)

        assert queue_status.number_of_pending_messages == 1
        assert queue_status.number_of_consumers == 0
        assert queue_status.messages_dequeued == 0


def test_should_publish_to_dql_due_to_implicit_nack_given_internal_callback_exception():
    test_destination_dlq_two = f"/queue/my-destination-dql-two-{uuid.uuid4()}"
    *_, queue_name = test_destination_dlq_two.split("/")
    dlq_queue_name = f"DLQ.{queue_name}"

    # In order to publish sample data
    some_body = {"keyOne": 1, "keyTwo": 2}
    publish_to_destination(test_destination_dlq_two, some_body)

    start_processing(test_destination_dlq_two, callback_with_exception_path, is_testing=True)

    try:
        queue_status = current_queue_configuration(dlq_queue_name)

        assert queue_status.number_of_pending_messages == 1
        assert queue_status.number_of_consumers == 0
        assert queue_status.messages_enqueued == 1
        assert queue_status.messages_dequeued == 0
    except Exception:
        queue_status = rabbitmq.current_queue_configuration(dlq_queue_name)

        assert queue_status.number_of_pending_messages == 1
        assert queue_status.number_of_consumers == 0
        assert queue_status.messages_dequeued == 0


def test_should_save_tshoot_properties_on_header():
    some_destination = f"tshoot-header-{uuid.uuid4()}"

    # In order to publish sample data
    some_body = {"keyOne": 1, "keyTwo": 2}
    publish_to_destination(some_destination, some_body)

    try:
        message_status = retrieve_message_published(some_destination)
    except Exception:
        message_status = rabbitmq.retrieve_message_published(some_destination)

    assert message_status.properties.get("tshoot-destination")
    assert message_status.properties["tshoot-destination"] == some_destination


def test_should_send_to_another_destination(caplog):
    caplog.set_level(logging.DEBUG)

    some_source_destination = f"/queue/source-{uuid.uuid4()}"
    some_target_destination = f"/queue/target-{uuid.uuid4()}"

    # In order to publish sample data
    with build_publisher().auto_open_close_connection() as publisher:
        some_body = {"keyOne": 1, "keyTwo": 2}
        some_headers = {"some-header-1": 1, "some-header-2": 2}
        publisher.send(some_body, some_source_destination, some_headers, attempt=1)

    move_messages_consumer = send_message_from_one_destination_to_another(
        some_source_destination, some_target_destination, is_testing=True, return_listener=True
    )

    wait_for_message_in_log(caplog, r"The messages has been moved")
    move_messages_consumer.close()

    *_, source_queue_name = some_source_destination.split("/")
    *_, target_queue_name = some_target_destination.split("/")
    try:
        source_queue_status = current_queue_configuration(source_queue_name)
        target_queue_status = current_queue_configuration(target_queue_name)
        message_status = retrieve_message_published(target_queue_name)
    except Exception:
        source_queue_status = rabbitmq.current_queue_configuration(source_queue_name)
        target_queue_status = rabbitmq.current_queue_configuration(target_queue_name)
        message_status = rabbitmq.retrieve_message_published(target_queue_name)

    assert source_queue_status.number_of_pending_messages == 0
    assert source_queue_status.number_of_consumers == 0
    assert source_queue_status.messages_enqueued == 1
    assert source_queue_status.messages_dequeued == 1

    assert target_queue_status.number_of_pending_messages == 1
    assert target_queue_status.number_of_consumers == 0
    assert target_queue_status.messages_enqueued == 1
    assert target_queue_status.messages_dequeued == 0

    keys = list(some_headers.keys())
    assert len(keys) == 2
    assert int(message_status.properties[keys[0]]) == some_headers[keys[0]]
    assert int(message_status.properties[keys[1]]) == some_headers[keys[1]]
    assert message_status.details == some_body


@mock.patch("django_stomp.execution.should_process_msg_on_background", False)
def test_should_use_heartbeat_and_then_lost_connection_due_message_takes_longer_than_heartbeat_no_background_processing(
    caplog, settings, mocker
):
    """
    The listener in this code has an, arguably broken, message handler (see
    _test_callback_function_with_sleep_three_seconds_while_heartbeat_thread_is_alive function) which awaits
    (blocking the consumer for three seconds) for a heartbeat timeout that occurs if no message is shared
    between the broker and the consumer in 1 second (1000 ms).

    Heartbeating requires an error margin due to timing inaccuracies which are usually configured
    by the receiver (such as the broker and even the client). Thus, other brokers may wait for a little bit more
    or less before considering the connection dead.
    """
    caplog.set_level(logging.DEBUG)
    some_destination = f"some-lorem-destination-{uuid.uuid4()}"

    # In order to publish sample data
    some_body = {"keyOne": 1, "keyTwo": 2}
    publish_to_destination(some_destination, some_body)

    settings.STOMP_OUTGOING_HEARTBEAT = 1000
    settings.STOMP_INCOMING_HEARTBEAT = 1000

    message_consumer = start_processing(
        some_destination,
        callback_with_sleep_three_seconds_while_heartbeat_thread_is_alive_path,
        is_testing=True,
        return_listener=True,
    )

    wait_for_message_in_log(caplog, r"Heartbeat loop ended", message_count_to_wait=2)
    message_consumer.close()

    assert any(
        re.compile(
            f"Received frame.+eart-beat.+{settings.STOMP_OUTGOING_HEARTBEAT},{settings.STOMP_INCOMING_HEARTBEAT}"
        ).match(m)
        for m in caplog.messages
    )
    assert any(re.compile("Sending a heartbeat message at [0-9.]+").match(message) for message in caplog.messages)
    assert any(
        re.compile("Heartbeat timeout: diff_receive=[0-9.]+, time=[0-9.]+, lastrec=[0-9.]+").match(message)
        for message in caplog.messages
    )
    assert any(re.compile("Heartbeat loop ended").match(message) for message in caplog.messages)


def test_should_create_queues_for_virtual_topic_listeners_and_consume_its_messages(caplog):
    caplog.set_level(logging.DEBUG)
    number_of_consumers = 2

    some_virtual_topic = f"VirtualTopic.{uuid.uuid4()}"
    consumers = [
        start_processing(
            f"Consumer.{uuid.uuid4()}.{some_virtual_topic}",
            callback_standard_path,
            is_testing=True,
            return_listener=True,
        )
        for _ in range(number_of_consumers)
    ]

    some_body = {"send": 2, "Virtual": "Topic"}
    publish_to_destination(f"/topic/{some_virtual_topic}", some_body)

    wait_for_message_in_log(
        caplog, r"Received frame: 'MESSAGE'.*body='{\"send\": 2, \"Virtual\": \"Topic\"}.*'", message_count_to_wait=2
    )

    for consumer in consumers:
        consumer.close()

    received_frame_log_messages = [
        message
        for message in caplog.messages
        if re.compile(r"Received frame: 'MESSAGE'.*body='{\"send\": 2, \"Virtual\": \"Topic\"}.*'").match(message)
    ]

    sending_frame_log_messages = [
        message for message in caplog.messages if re.compile(r"Sending frame: \[b'ACK'.*").match(message)
    ]

    assert len(received_frame_log_messages) == number_of_consumers
    assert len(sending_frame_log_messages) == number_of_consumers


def test_should_create_queue_for_virtual_topic_consumer_and_process_messages_sent_when_consumer_is_disconnected(caplog):
    caplog.set_level(logging.DEBUG)

    some_virtual_topic = f"VirtualTopic.{uuid.uuid4()}"
    virtual_topic_consumer_queue = f"Consumer.{uuid.uuid4()}.{some_virtual_topic}"
    consumer = start_processing(
        virtual_topic_consumer_queue, callback_with_logging_path, is_testing=True, return_listener=True
    )

    some_body = {"send": 2, "another": "VirtualTopic"}
    publish_to_destination(f"/topic/{some_virtual_topic}", some_body)

    wait_for_message_in_log(caplog, r"I'll process the message: {'send': 2, 'another': 'VirtualTopic'}!")
    consumer.close()

    # Send message to VirtualTopic with consumers disconnected...
    some_body = {"send": "anotherMessage", "2": "VirtualTopic"}
    publish_to_destination(f"/topic/{some_virtual_topic}", some_body)

    # Start another consumer for the same queue created previously...
    another_consumer = start_processing(
        virtual_topic_consumer_queue, callback_with_logging_path, is_testing=True, return_listener=True
    )

    wait_for_message_in_log(caplog, r"I'll process the message: {'send': 'anotherMessage', '2': 'VirtualTopic'}!")
    another_consumer.close()

    received_first_frame_log_messages = [
        message
        for message in caplog.messages
        if re.compile(r"I'll process the message: {'send': 2, 'another': 'VirtualTopic'}!").match(message)
    ]
    received_second_frame_log_messages = [
        message
        for message in caplog.messages
        if re.compile(r"I'll process the message: {'send': 'anotherMessage', '2': 'VirtualTopic'}!").match(message)
    ]

    sending_frame_log_messages = [
        message for message in caplog.messages if re.compile(r"Sending frame: \[b'ACK'.*").match(message)
    ]

    assert len(received_first_frame_log_messages) == 1
    assert len(received_second_frame_log_messages) == 1
    assert len(sending_frame_log_messages) == 2


def test_should_create_queue_for_virtual_topic_and_compete_for_its_messages(caplog):
    caplog.set_level(logging.DEBUG)

    some_virtual_topic = f"VirtualTopic.{uuid.uuid4()}"
    virtual_topic_consumer_queue = f"Consumer.{uuid.uuid4()}.{some_virtual_topic}"

    consumers = [
        start_processing(
            virtual_topic_consumer_queue, callback_with_logging_path, is_testing=True, return_listener=True
        ),
        start_processing(
            virtual_topic_consumer_queue, callback_with_another_log_message_path, is_testing=True, return_listener=True
        ),
    ]

    with build_publisher().auto_open_close_connection() as publisher:
        with publisher.do_inside_transaction():
            some_body = {"send": 2, "some": "VirtualTopic"}
            publisher.send(some_body, f"/topic/{some_virtual_topic}", attempt=1)
            publisher.send(some_body, f"/topic/{some_virtual_topic}", attempt=1)
            publisher.send(some_body, f"/topic/{some_virtual_topic}", attempt=1)

    wait_for_message_in_log(caplog, r"Sending frame: \[b'ACK'.*", message_count_to_wait=3)

    for consumer in consumers:
        consumer.close()

    callback_logs = [
        message
        for message in caplog.messages
        if re.compile(r"I'll process the message: {'send': 2, 'some': 'VirtualTopic'}!").match(message)
    ]
    another_callback_logs = [
        message
        for message in caplog.messages
        if re.compile(r"{'send': 2, 'some': 'VirtualTopic'} is the message that I'll process!").match(message)
    ]

    sending_frame_log_messages = [
        message for message in caplog.messages if re.compile(r"Sending frame: \[b'ACK'.*").match(message)
    ]

    assert 1 <= len(callback_logs) <= 2
    assert 1 <= len(another_callback_logs) <= 2
    assert len(callback_logs + another_callback_logs) == 3
    assert len(sending_frame_log_messages) == 3


def test_shouldnt_process_message_from_virtual_topic_older_than_the_consumer_queue_creation():
    some_virtual_topic = f"VirtualTopic.{uuid.uuid4()}"
    virtual_topic_consumer_queue = f"Consumer.{uuid.uuid4()}.{some_virtual_topic}"

    with build_publisher().auto_open_close_connection() as publisher:
        some_body = {"send": 2, "topicThat": "isVirtual"}
        publisher.send(some_body, f"/topic/{some_virtual_topic}", attempt=1)

    start_processing(virtual_topic_consumer_queue, callback_with_logging_path, is_testing=True)

    try:
        queue_status = current_queue_configuration(virtual_topic_consumer_queue)

        assert queue_status.number_of_pending_messages == 0
        assert queue_status.number_of_consumers == 0
        assert queue_status.messages_enqueued == 0
        assert queue_status.messages_dequeued == 0
    except Exception:
        queue_status = rabbitmq.current_queue_configuration(virtual_topic_consumer_queue)

        assert queue_status.number_of_pending_messages == 0
        assert queue_status.number_of_consumers == 0
        assert queue_status.messages_dequeued == 0


@mock.patch("django_stomp.execution.is_correlation_id_required", True)
def test_should_raise_exception_when_correlation_id_is_required_but_not_received(settings):
    test_destination_one = f"/queue/test-correlation-id-required-one-{uuid4()}"
    test_destination_two = f"/queue/test-correlation-id-required-two-{uuid4()}"

    some_body = {"keyOne": 1, "keyTwo": 2}
    publish_without_correlation_id_header(test_destination_one, some_body)

    with pytest.raises(Exception):
        assert start_processing(
            test_destination_one, callback_move_and_ack_path, param_to_callback=test_destination_two, is_testing=True
        )


@mock.patch("django_stomp.execution.is_correlation_id_required", True)
def test_should_raise_exception_when_correlation_id_is_not_supplied_and_publish_it_to_dlq(settings, caplog):
    some_body = {"keyOne": 1, "keyTwo": 2}
    destination_name = f"/queue/destination-for-correlation-id-testing-required-{uuid4()}"

    # removing /queue/
    *_, source_queue_name = destination_name.split("/")
    dlq_source_queue_name = f"DLQ.{source_queue_name}"

    # publishes messages WITHOUT correlation-id header, but message must be persistent or won't go to DLQ
    # https://activemq.apache.org/message-redelivery-and-dlq-handling
    publish_without_correlation_id_header(destination_name, some_body, persistent=True)

    consumer = start_processing(destination_name, callback_standard_path, is_testing=True, return_listener=True)
    wait_for_message_in_log(
        caplog,
        r"A exception of type <class 'django_stomp\.exceptions\.CorrelationIdNotProvidedException'>.*",
        message_count_to_wait=1,
    )
    consumer.close()

    dlq_source_queue_status = get_destination_metrics_from_broker(dlq_source_queue_name)

    # asserts that DLQ has a message on it!
    assert dlq_source_queue_status.number_of_pending_messages == 1
    assert dlq_source_queue_status.number_of_consumers == 0


@mock.patch("django_stomp.execution.is_correlation_id_required", False)
def test_should_consume_message_without_correlation_id_when_it_is_not_required_no_dlq(settings):
    some_body = {"keyOne": 1, "keyTwo": 2}
    destination_name = f"/queue/destination-for-correlation-id-testing-not-required-{uuid4()}"

    # removing /queue/
    *_, source_queue_name = destination_name.split("/")
    dlq_source_queue_name = f"DLQ.{source_queue_name}"

    # publishes messages WITHOUT correlation-id header, but message must be persistent or won't go to DLQ
    # https://activemq.apache.org/message-redelivery-and-dlq-handling
    publish_without_correlation_id_header(destination_name, some_body, persistent=True)

    start_processing(destination_name, callback_standard_path, is_testing=True)

    source_queue_status = get_destination_metrics_from_broker(source_queue_name)

    # normal queue should have ONE dequeued message: no exception was raised!
    assert source_queue_status.number_of_pending_messages == 0
    assert source_queue_status.number_of_consumers == 0
    assert source_queue_status.messages_dequeued == 1

    dlq_source_queue_status = get_destination_metrics_from_broker(dlq_source_queue_name)

    # DLQ should have NO messages
    assert dlq_source_queue_status.number_of_pending_messages == 0
    assert dlq_source_queue_status.number_of_consumers == 0
    assert dlq_source_queue_status.messages_dequeued == 0


@mock.patch("django_stomp.execution.should_process_msg_on_background", True)
def test_should_use_heartbeat_and_dont_lose_connection_when_using_background_processing(caplog, settings, mocker):
    """
    The listener in this code has a message handler that takes a long time to process a message (see
    _test_callback_function_with_sleep_three_seconds function). Using a background process to process
    the message, the heartbeat functionality should still work, maintaining the connection
    estabilished between broker and consumer.
    """
    caplog.set_level(logging.DEBUG)
    some_destination = f"some-lorem-destination-{uuid.uuid4()}"

    # In order to publish sample data
    some_body = {"keyOne": 1, "keyTwo": 2}
    publish_to_destination(some_destination, some_body)

    settings.STOMP_OUTGOING_HEARTBEAT = 1000
    settings.STOMP_INCOMING_HEARTBEAT = 1000

    message_consumer = start_processing(
        some_destination, callback_with_sleep_three_seconds_path, is_testing=True, return_listener=True
    )

    wait_for_message_in_log(caplog, f"{some_body} sucessfully processed!", message_count_to_wait=1)
    message_consumer.close()

    received_heartbeat_frame_regex = re.compile(
        f"Received frame.+heart-beat.+{settings.STOMP_OUTGOING_HEARTBEAT},{settings.STOMP_INCOMING_HEARTBEAT}"
    )
    sending_heartbeat_message_regex = re.compile("Sending a heartbeat message at [0-9.]+")
    sending_ack_frame_regex = re.compile(f"Sending frame: .+ACK.+subscription:{message_consumer._subscription_id}.+")
    heartbeat_timeout_regex = re.compile("Heartbeat timeout: diff_receive=[0-9.]+, time=[0-9.]+, lastrec=[0-9.]+")

    assert any(received_heartbeat_frame_regex.match(m) for m in caplog.messages)
    assert any(sending_heartbeat_message_regex.match(m) for m in caplog.messages)
    assert any(sending_ack_frame_regex.match(m) for m in caplog.messages)
    assert not any(heartbeat_timeout_regex.match(m) for m in caplog.messages)


@pytest.mark.skipif(is_testing_against_rabbitmq(), reason="RabbitMQ doesn't holds the concept of a durable subscriber")
def test_shouldnt_create_a_durable_subscriber_when_dealing_with_virtual_topics():
    all_offline_subscribers_before_the_virtual_topic_connection = list(offline_durable_subscribers("localhost"))
    destination_two = f"/queue/testing-for-durable-subscriber-with-virtual-topics-{uuid4()}"

    some_virtual_topic = f"VirtualTopic.{uuid.uuid4()}"
    virtual_topic_consumer_queue = f"Consumer.{uuid.uuid4()}.{some_virtual_topic}"
    start_processing(
        virtual_topic_consumer_queue, callback_move_and_ack_path, param_to_callback=destination_two, is_testing=True
    )

    all_offline_subscribers_after_the_virtual_topic_connection = list(offline_durable_subscribers("localhost"))

    assert sorted(all_offline_subscribers_before_the_virtual_topic_connection) == sorted(
        all_offline_subscribers_after_the_virtual_topic_connection
    )


def test_should_connect_with_a_queue_created_without_the_durable_header(caplog):
    caplog.set_level(logging.DEBUG)
    some_destination = f"/queue/some-queue-without-durable-headers-{uuid.uuid4()}"

    listener = build_listener(some_destination)
    listener._subscription_configuration.pop("durable")
    listener.start(wait_forever=False)
    listener.close()

    send_frames_with_durable_true_regex = re.compile(r"^Sending frame:.*durable:true.*")

    assert all(not send_frames_with_durable_true_regex.match(m) for m in caplog.messages)

    some_header = {"correlation-id": uuid.uuid4()}
    some_body = {"keyOne": 1, "keyTwo": 2}
    publish_to_destination(some_destination, some_body, some_header)

    # Calling what we need to test
    message_consumer = start_processing(
        some_destination, callback_with_logging_path, is_testing=True, return_listener=True
    )

    wait_for_message_in_log(caplog, r"Sending frame: \[b'ACK'.*", message_count_to_wait=1)
    message_consumer.close()

    consumer_log_message_regex = re.compile(f"I'll process the message: {some_body}!")

    assert any(send_frames_with_durable_true_regex.match(m) for m in caplog.messages)
    assert any(consumer_log_message_regex.match(m) for m in caplog.messages)


def test_should_clean_all_messages_on_a_destination(caplog):
    caplog.set_level(logging.DEBUG)
    trash_msgs_count = 10
    some_source_destination = f"/queue/flood-this-queue-with-trash-{uuid4()}"
    *_, source_queue_name = some_source_destination.split("/")

    some_body = {"some": "trash"}
    some_headers = {"some": "header"}

    # command invocation
    consumer = clean_messages_on_destination_by_acking(some_source_destination, is_testing=True, return_listener=True)

    with build_publisher().auto_open_close_connection() as publisher:
        for _ in range(0, trash_msgs_count):
            publisher.send(some_body, some_source_destination, some_headers, attempt=1)

    wait_for_message_in_log(caplog, r"Message has been removed!", message_count_to_wait=trash_msgs_count)
    consumer.close()

    # asserts that messages were acked
    source_queue_status = get_destination_metrics_from_broker(source_queue_name)

    assert source_queue_status.number_of_pending_messages == 0
    assert source_queue_status.number_of_consumers == 0
    assert source_queue_status.messages_enqueued == trash_msgs_count  # enqueued the queue with trash!
    assert source_queue_status.messages_dequeued == trash_msgs_count  # cleaned all the trash on the queue!


@mock.patch("django_stomp.execution.is_correlation_id_required", True)
def test_should_clean_problematic_headers_and_publish_it_to_destination_successfully(settings, caplog):
    destination_name = f"/queue/headers-cleaning-destination-{uuid4()}"
    *_, source_queue_name = destination_name.split("/")

    loyalty_event_body = {
        "points": 300,
        "account": "7cdd43af-7e36-40c2-b7e7-e0e0a62fa328",
        "sponsor": "3591bfff-3241-46e3-8829-2389d56da04c",
    }

    dangerous_headers = {
        "subscription": "77625173-42b5-4c62-9c24-1fc4a4668dcc-listener",
        "destination": "/queue/wrong-destination",
        "message-id": "T_77625173-42b5-4c62-9c24-1fc4a4668dcc-listener@@session-jSfP4eyH59eKxHaZDv59Cg@@2",
        "redelivered": "false",
        "x-dead-letter-routing-key": "DLQ.wrong-destination",
        "x-dead-letter-exchange": "",
        "tshoot-destination": "/queue/wrong-destination",
        "transaction": "11837f0b-f1c5-494f-9b2b-347c636d8ca2",
        "eventType": "someType",
        "correlation-id": "6e10903e13ddcde54304883e5df713a5",
        "persistent": "true",
        "content-type": "application/json;charset=utf-8",
        "content-length": "298",
    }

    # headers such as message-id, transaction are reserved for internal use of RabbitMQ: they should
    # generate problems if NOT removed
    publish_to_destination(destination_name, loyalty_event_body, dangerous_headers)

    # published message assertions
    received_message = get_latest_message_from_destination_using_test_listener(destination_name)
    message_headers = received_message[0]
    message_body = json.loads(received_message[1])

    assert message_headers["tshoot-destination"] == destination_name
    assert message_headers["x-dead-letter-routing-key"] == f"DLQ.{source_queue_name}"
    assert message_headers["correlation-id"] == "6e10903e13ddcde54304883e5df713a5"
    assert message_headers["eventType"] == "someType"

    assert message_body["points"] == 300
    assert message_body["account"] == "7cdd43af-7e36-40c2-b7e7-e0e0a62fa328"
    assert message_body["sponsor"] == "3591bfff-3241-46e3-8829-2389d56da04c"


def test_should_not_publish_any_messages_if_connection_drops_when_using_transactions():
    # Base environment setup
    destination_one = f"/queue/no-retrying-destination-one-{uuid4()}"
    *_, queue_name = destination_one.split("/")
    some_correlation_id = uuid.uuid4()
    some_header = {"correlation-id": some_correlation_id}
    some_body = {"please": "no errors"}

    # creates destination and publishes to it
    start_processing(destination_one, callback_standard_path, is_testing=True, return_listener=True).close()
    publisher = build_publisher(f"random-publisher-{uuid4()}")

    with pytest.raises(NotConnectedException):
        # transaction with connection errors: everything should be aborted and no messages published
        with do_inside_transaction(publisher):
            publisher.send(some_body, queue=destination_one, headers=some_header)
            publisher.close()  # simulates a closed connection/timeout/broken pipe/etc IN A TRANSACTION
            publisher.send(some_body, queue=destination_one, headers=some_header)

    queue_status = get_destination_metrics_from_broker(queue_name)

    # no messages should have been published ahead
    assert queue_status.number_of_pending_messages == 0


def test_should_publish_many_messages_if_no_connection_problems_happen_when_using_transactions():
    # Base environment setup
    destination_one = f"/queue/no-retrying-destination-one-{uuid4()}"
    *_, queue_name = destination_one.split("/")
    some_correlation_id = uuid.uuid4()
    some_header = {"correlation-id": some_correlation_id}
    some_body = {"please": "no errors"}

    # creates destination and publishes to it
    start_processing(destination_one, callback_standard_path, is_testing=True, return_listener=True).close()
    publisher = build_publisher(f"random-publisher-{uuid4()}")

    # no connection errors inside this transaction
    with do_inside_transaction(publisher):
        publisher.send(some_body, queue=destination_one, headers=some_header)
        publisher.send(some_body, queue=destination_one, headers=some_header)

    queue_status = get_destination_metrics_from_broker(queue_name)

    # two messages should have been published (no connection/errors on send)
    assert queue_status.messages_enqueued == 2
    assert queue_status.number_of_pending_messages == 2


def test_should_publish_messages_if_connection_drops_when_not_transactions():
    # Base environment setup
    destination_one = f"/queue/yes-retrying-destination-one-{uuid4()}"
    *_, queue_name = destination_one.split("/")
    some_correlation_id = uuid.uuid4()
    some_header = {"correlation-id": some_correlation_id}
    some_body = {"please": "no errors"}

    # creates destination and publishes to it
    start_processing(destination_one, callback_standard_path, is_testing=True, return_listener=True).close()
    publisher = build_publisher(f"random-publisher-{uuid4()}")

    # no transaction
    publisher.send(some_body, queue=destination_one, headers=some_header)
    publisher.close()  # simulates a closed connection/timeout/broken pipe/etc
    publisher.send(some_body, queue=destination_one, headers=some_header)

    # some sleep to give some time for RabbitMQ Console
    sleep(2)
    queue_status = get_destination_metrics_from_broker(queue_name)

    # the two message should have been published due to retry (but no transactions involved)
    assert queue_status.messages_enqueued == 2
    assert queue_status.number_of_pending_messages == 2


@pytest.mark.django_db
def test_should_open_a_new_db_connection_when_previous_connection_is_obsolete_or_unusable(settings):
    # Arrange - listen for db connection created signal with mocked function
    new_db_connection_callback = mock.MagicMock(return_value=None)
    connection_created.connect(new_db_connection_callback)

    # Arrange - settings CONN_MAX_AGE to zero, forcing connection renewal for every new message
    settings.DATABASES["default"]["CONN_MAX_AGE"] = 0

    # Arrange - random destination and fake arbitrary message body
    destination = f"/queue/new-connections-{uuid4()}"
    body = {"key": "value"}
    arbitrary_number_of_msgs = 5

    # Act - start listening to a random destination queue and publishing an arbitrary number of messages to it
    listener = start_processing(
        destination, callback_with_explicit_db_connection_path, is_testing=True, return_listener=True
    )

    with build_publisher(f"random-publisher-{uuid4()}").auto_open_close_connection() as publisher:
        for _ in range(arbitrary_number_of_msgs):
            publisher.send(body, destination)

    sleep(0.5)  # some sleep to give enough time to process the messages

    # Assert - for every new message a new db connection is created
    assert new_db_connection_callback.call_count == arbitrary_number_of_msgs

    # Assert - all threads that have established a db connection, should have closed them as CONN_MAX_AGE == 0
    threads_with_db_connections = [t for t in threading.enumerate() if hasattr(t, "db")]
    assert any(threads_with_db_connections)
    assert all(t.db.connection is None for t in threads_with_db_connections)

    listener.close()


@pytest.mark.django_db
def test_shouldnt_open_a_new_db_connection_when_there_is_one_still_usable(settings):
    # Arrange - listen for db connection created signal with mocked function
    new_db_connection_callback = mock.MagicMock(return_value=None)
    connection_created.connect(new_db_connection_callback)

    # Arrange - settings CONN_MAX_AGE to 86400s or 1 day
    settings.DATABASES["default"]["CONN_MAX_AGE"] = 86400

    # Arrange - random destination and fake arbitrary message body
    destination = f"/queue/no-new-connections-{uuid4()}"
    body = {"key": "value"}
    arbitrary_number_of_msgs = 5

    # Act - start listening random destination queue and publish arbitrary number of messages to it
    listener = start_processing(
        destination, callback_with_explicit_db_connection_path, is_testing=True, return_listener=True
    )

    with build_publisher(f"random-publisher-{uuid4()}").auto_open_close_connection() as publisher:
        for _ in range(arbitrary_number_of_msgs):
            publisher.send(body, destination)

    sleep(0.5)  # some sleep to give enough time to process the messages

    # Assert - only one connection is estabilished
    assert new_db_connection_callback.call_count == 1

    # Assert - all threads that have established a db connection, shouldn't have reset them as CONN_MAX_AGE == 1 day
    threads_with_db_connections = [t for t in threading.enumerate() if hasattr(t, "db")]
    assert any(threads_with_db_connections)
    assert all(t.db.connection is not None for t in threads_with_db_connections)

    listener.close()



from time import sleep
from typing import Generator

import requests

from django_stomp.helpers import eval_str_as_boolean
from parsel import Selector
from tests.support.dtos import ConsumerStatus


def consumers_details(connection_id, host="localhost") -> Generator[ConsumerStatus, None, None]:
    sleep(1)

    params = {"connectionID": connection_id}
    result = requests.get(f"http://{host}:8161/admin/connection.jsp", params=params, auth=("admin", "admin"))
    selector = Selector(text=str(result.content))

    consumers_table = selector.xpath('//*[@id="messages"]/tbody/tr').getall()

    assert len(consumers_table) > 0
    for index, consumer_details in enumerate(consumers_table):
        consumer_details_as_selector = Selector(text=consumer_details)
        destination_request_path = consumer_details_as_selector.css("td a::attr(href)").get()

        address_to_destination_details = f"http://{host}:8161/admin/{destination_request_path}"
        destination_name = destination_request_path.split("Destination=")[1]
        session_id = int(consumer_details_as_selector.css("td + td::text").get())
        enqueues = int(consumer_details_as_selector.css("td + td + td::text").get())
        dequeues = int(consumer_details_as_selector.css("td + td + td + td::text").get())
        dispatched = int(consumer_details_as_selector.css("td + td + td + td + td::text").get())
        dispatched_queue = int(consumer_details_as_selector.css("td + td + td + td + td + td::text").get())

        last_two_columns = consumer_details_as_selector.css("td + td + td + td + td + td+ td + td::text").getall()

        prefetch = int(last_two_columns[0].replace("\\", "").replace("t", "").replace("n", ""))
        max_pending = int(last_two_columns[1].replace("\\", "").replace("t", "").replace("n", ""))
        exclusive = eval_str_as_boolean(last_two_columns[2].replace("\\", "").replace("t", "").replace("n", ""))
        retroactive = eval_str_as_boolean(last_two_columns[3].replace("\\", "").replace("t", "").replace("n", ""))

        yield ConsumerStatus(
            address_to_destination_details,
            destination_name,
            session_id,
            enqueues,
            dequeues,
            dispatched,
            dispatched_queue,
            prefetch,
            max_pending,
            exclusive,
            retroactive,
        )



from django.core.management.base import BaseCommand
from django_stomp.exceptions import DjangoStompIncorrectUse
from django_stomp.execution import clean_messages_on_destination_by_acking


class Command(BaseCommand):
    help = "Cleans a destination by acking all messages on it!"

    def add_arguments(self, parser):
        parser.add_argument("destination_to_clean", type=str, help="Destination to be cleaned from all messages")

    def handle(self, *args, **options):
        self.stdout.write(f"Provided parameters: {options}")
        source_destination = options.get("destination_to_clean")

        if not source_destination:
            raise DjangoStompIncorrectUse("No destination_to_clean was supplied! Nothing to clean!")

        self.stdout.write(f"Preparing to clean the queue: {source_destination}")
        clean_messages_on_destination_by_acking(source_destination)



from io import StringIO

import pytest
from django.core.management import CommandError
from django.core.management import call_command
from pytest_mock import MockFixture


def test_should_call_start_processing(mocker: MockFixture):
    mock_start_processing = mocker.patch("django_stomp.management.commands.pubsub.start_processing")
    fake_queue = "/queue/your-queue"
    fake_function = "your_python_module.your_function"
    out = StringIO()

    call_command("pubsub", fake_queue, fake_function, stdout=out)

    assert "Calling internal service to consume messages" in out.getvalue()
    mock_start_processing.assert_called_with(fake_queue, fake_function)


def test_should_essential_parameters_are_required(mocker: MockFixture):
    mock_start_processing = mocker.patch("django_stomp.management.commands.pubsub.start_processing")
    fake_queue = "/queue/your-queue"

    with pytest.raises(CommandError) as e:
        call_command("pubsub")

    assert e.value.args[0] == "Error: the following arguments are required: source_destination, callback_function"

    with pytest.raises(CommandError) as e:
        call_command("pubsub", fake_queue)

    assert e.value.args[0] == "Error: the following arguments are required: callback_function"

    mock_start_processing.assert_not_called()



import json
from time import sleep

import requests

from parsel import Selector
from tests.support.dtos import MessageStatus


def retrieve_message_published(destination_name, host="localhost") -> MessageStatus:
    sleep(1)
    address = f"http://{host}:8161/admin/browse.jsp?JMSDestination={destination_name}"
    result = requests.get(address, auth=("admin", "admin"))
    selector = Selector(text=str(result.content))

    all_messages = selector.xpath('//*[@id="messages"]/tbody/tr').getall()

    assert len(all_messages) > 0

    for index, message_details in enumerate(all_messages):
        message_details_as_selector = Selector(text=message_details)
        message_id_request_path = message_details_as_selector.css("td a::attr(href)").get()
        address_to_message = f"http://{host}:8161/admin/{message_id_request_path}"
        result_all_message_details = requests.get(address_to_message, auth=("admin", "admin"))
        selector_all_message_details = Selector(text=str(result_all_message_details.content))
        return _process_message_details(selector_all_message_details)


def _process_message_details(selector_message_details) -> MessageStatus:
    message_id = selector_message_details.css("table#header td + td::text").get()
    details = json.loads(selector_message_details.css("pre::text").get())
    css_persistent_location = "table#header tr + tr + tr + tr + tr + tr + tr td + td::text"
    persistent = "persistent" in selector_message_details.css(css_persistent_location).get().lower()
    correlation_id = selector_message_details.css("table#header tr + tr + tr td + td::text").get()

    properties = {}
    all_configured_properties = selector_message_details.css("table#properties td").getall()
    for index, configured_property in enumerate(all_configured_properties):
        configured_property_as_selector = Selector(text=configured_property)
        key = configured_property_as_selector.css("td[class='label']::text").get()
        if key:
            value = Selector(text=all_configured_properties[index + 1]).css("td::text").get()
            properties.update({key: value})

    return MessageStatus(message_id, details, persistent, correlation_id, properties)
